"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np
from typing import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.optimization.dv_output_defs import *
from cached_property import cached_property
from adsg_core.func_cache import cached_function, clear_func_cache
from adsg_core.graph.adsg import DSGType
from adsg_core.optimization.hierarchy import *
from adsg_core.optimization.assign_enc.encoding import Encoder
from adsg_core.optimization.assign_enc.selector import EncoderSelector
from adsg_core.optimization.assign_enc.time_limiter import run_timeout
from adsg_core.optimization.assign_enc.assignment_manager import AssignmentManagerBase

__all__ = ['GraphProcessor', 'MetricType', 'SelChoiceEncoderType']


def catch_memory_overflow(func):
    def wrapper(obj: 'GraphProcessor', *args, **kwargs):
        try:
            return func(obj, *args, **kwargs)

        except MemoryError:
            # Enable memory save mode (if not enabled yet) and try again
            if not obj._memory_save_mode:
                obj._memory_save_mode = True
                clear_func_cache(obj)
                return wrapper(obj, *args, **kwargs)

            else:
                raise

    wrapper.__name__ = func.__name__
    return wrapper


class GraphProcessor:
    """
    Base class for the graph processor: the object that captures all logic for using an architecture graph in an
    optimization problem. A graph processor specifies the optimization problem parameters (design variables, objectives,
    constraints, other output) and contains methods for converting design variables to objectives/constraints/output.

    The graph passed to the processor should be one describing the whole design space, and it should already have
    external functions defined.
    """

    default_encoding_timeout = 10  # sec
    encoders: Dict[SelChoiceEncoderType, Type[HierarchyAnalyzerBase]] = {
        # SelChoiceEncoderType.COMPLETE: SelChoiceEncHierarchyAnalyzer,
        SelChoiceEncoderType.COMPLETE: HierarchyAnalyzer,
        SelChoiceEncoderType.FAST: FastHierarchyAnalyzer,
    }
    _n_combs_cutoff = 1e9

    def __init__(self, graph: DSGType, encoding_timeout: float = None, encoder_type: SelChoiceEncoderType = None):
        self._graph: DSGType = self._check_graph(graph)
        self._fixed_values: Dict[int, Union[int, float]] = {}
        self._comb_fixed_mask = None
        self._graph_cache = {}
        self._excluded_cache = set()
        self.encoding_timeout = encoding_timeout or self.default_encoding_timeout
        self._encoder_type = encoder_type
        self._reset_conn_encoder_cache = False
        self._memory_save_mode = False

    @property
    def graph(self) -> DSGType:
        """The DSG that forms the basis for this optimization problem"""
        return self._graph

    @cached_property
    def _hierarchy_analyzer(self) -> HierarchyAnalyzerBase:
        return self._get_hierarchy_analyzer()

    def _get_hierarchy_analyzer(self) -> HierarchyAnalyzerBase:
        if self._encoder_type is not None:
            return self.encoders[self._encoder_type](self._graph)

        def _do_encode():
            an = self.encoders[SelChoiceEncoderType.COMPLETE](self._graph)

            # Trigger encoding
            assert an.n_combinations >= 0

            return an

        try:
            analyzer = run_timeout(self.encoding_timeout, _do_encode)

        except (TimeoutError, MemoryError):
            analyzer = self.encoders[SelChoiceEncoderType.FAST](self._graph)

        return analyzer

    @property
    def encoder_type(self):
        return self._hierarchy_analyzer.get_encoder_type()

    @cached_property
    def selection_choice_nodes(self):
        return self._hierarchy_analyzer.selection_choice_nodes

    @cached_property
    def _sel_choice_opt_nodes(self):
        return self._hierarchy_analyzer.selection_choice_option_nodes

    @cached_property
    def _sel_choice_is_forced(self):
        return self._hierarchy_analyzer.selection_choice_is_forced

    @cached_property
    def _all_des_var_data(self):
        return self._get_des_vars()

    @property
    def all_des_vars(self) -> List[DesVar]:
        return self._all_des_var_data[0]

    @property
    def _sel_choice_idx_map(self):
        return self._all_des_var_data[1]

    @property
    def _conn_choice_data_map(self):
        return self._all_des_var_data[2]

    @property
    def _existence_infeasibility_mask(self):
        return self._all_des_var_data[3]

    @property
    def _existence_mask(self):
        if self._comb_fixed_mask is not None:
            return self._existence_infeasibility_mask & self._comb_fixed_mask
        return self._existence_infeasibility_mask

    @cached_property
    def all_des_var_idx_map(self):
        return {des_var: i for i, des_var in enumerate(self.all_des_vars)}

    @cached_property
    def all_des_var_nodes(self):
        return {des_var.node for des_var in self.all_des_vars}

    @cached_property
    def all_des_var_node_map(self):
        des_vars = self.all_des_vars
        return {node: [des_var for des_var in des_vars if des_var.node is node] for node in self.all_des_var_nodes}

    @property
    def des_vars(self) -> List[DesVar]:
        """The optimization problem design variables"""
        fixed_values = self._fixed_values
        return [des_var for i, des_var in enumerate(self.all_des_vars) if i not in fixed_values]

    @property
    def dv_is_conditionally_active(self) -> List[bool]:
        return [dv.conditionally_active for dv in self.des_vars]

    @property
    def fixed_values(self):
        return self._fixed_values.copy()

    @cached_property
    def objectives(self) -> List[Objective]:
        """The optimization problem objectives"""
        return self._categorized_metrics[0]

    @cached_property
    def constraints(self) -> List[Constraint]:
        """The optimization problem constraints"""
        return self._categorized_metrics[1]

    @cached_property
    def permanent_nodes(self):
        return self._get_permanent_nodes()

    @cached_property
    def _metrics(self):
        return self._get_metrics()

    @cached_property
    def _categorized_metrics(self):
        return self._categorize_metrics()

    @staticmethod
    def _check_graph(graph: DSGType) -> DSGType:
        """
        Check the initially provided graph:
        - Should have external functions defined
        - Should have choices to be made
        """

        # Check if start nodes have been defined
        if graph.derivation_start_nodes is None:
            raise ValueError('The provided graph should have derivation-start nodes defined')

        # # Check if there are any choices
        # if len(graph.get_nodes_by_type(SelectionChoiceNode)) == 0:
        #     if len(graph.choice_nodes) == 0 and len(graph.des_var_nodes) == 0:
        #         raise ValueError('The provided graph contains no choices: there is no optimization to be done!')

        # Check if the initial graph is feasible
        if not graph.feasible:
            raise ValueError('The provided graph is not feasible to begin with!')

        return graph

    @catch_memory_overflow
    def _get_des_vars(self) -> Tuple[List[DesVar], List[int], Dict[ConnectionChoiceNode, tuple], np.ndarray]:
        """
        The design variables for a given architecture graph are defined as follows (and in this order):
        - Each SelectionChoiceNode gets one DesVar
        - Each ConnectionChoiceNode gets one or more DesVars, depending on the encoding scheme
        - Each DesignVariableNode gets one DesVar
        """

        # Check if there are any feasible graphs to start with
        if self._hierarchy_analyzer.n_combinations == 0:
            raise RuntimeError('There are no feasible graphs to begin with!')

        # Create design variables for selection-choice nodes
        des_vars = []
        sel_choice_idx_map = []
        existing_names = set()
        sel_choice_opt_nodes = self._sel_choice_opt_nodes
        is_forced = self._sel_choice_is_forced
        permanent_nodes = self._hierarchy_analyzer.influence_matrix.permanent_nodes_incl_choice_nodes
        for i_dec, choice_node in enumerate(self.selection_choice_nodes):
            if is_forced[i_dec]:
                continue
            options = sel_choice_opt_nodes[choice_node]
            des_vars.append(DesVar.from_choice_node(choice_node, options, existing_names=existing_names,
                                                    conditionally_active=choice_node not in permanent_nodes))
            sel_choice_idx_map.append(i_dec)

            # Track existing names
            existing_names = {des_var.name for des_var in des_vars}

        # Connection choice nodes
        cutoff_mode = False
        if self._memory_save_mode:
            n_combs = 1
            cutoff_mode = True
        else:
            n_combs = self._hierarchy_analyzer.n_combinations
            if n_combs > self._n_combs_cutoff:
                n_combs = 1
                cutoff_mode = True

        existence_infeasibility_mask = np.ones((n_combs,), dtype=bool)
        conn_choice_data_map = {}
        for choice_node in self.connection_choice_nodes:
            assignment_manager, conn_des_vars, node_map, exist_map, all_conn_nodes = \
                self._encode_connection_choice(choice_node)
            i_dv_start = len(des_vars)

            for conn_des_var in conn_des_vars:
                if conn_des_var.node not in permanent_nodes:
                    conn_des_var.conditionally_active = True

            des_vars += conn_des_vars
            i_dv_end = len(des_vars)

            # Exclude selection-choice combinations that lead to infeasible port connections
            if not isinstance(exist_map, dict) and not cutoff_mode:
                existence_infeasibility_mask[exist_map == -1] = False

            conn_choice_data_map[choice_node] = \
                (assignment_manager, node_map, exist_map, i_dv_start, i_dv_end, all_conn_nodes)

        # Create design variables for continuous design variable nodes
        des_vars += [DesVar.from_des_var_node(des_var_node, conditionally_active=des_var_node not in permanent_nodes)
                     for des_var_node in self.design_variable_nodes]

        return des_vars, sel_choice_idx_map, conn_choice_data_map, existence_infeasibility_mask

    def fix_des_var(self, des_var: DesVar, value: Union[int, float] = None):
        """
        Fix a design variable to a specific value. Provide `None` to unfix.
        """
        idx = self.all_des_vars.index(des_var)
        if value is None:
            if idx in self._fixed_values:
                del self._fixed_values[idx]
        else:
            if des_var.is_discrete:
                if value < 0 or value >= des_var.n_opts:
                    raise ValueError('Value (%d) out of range: <%d' % (value, des_var.n_opts))
            else:
                if value < des_var.bounds[0] or value > des_var.bounds[1]:
                    raise ValueError('Value out of bounds: %.2f <= %.2f <= %.2f' %
                                     (des_var.bounds[0], value, des_var.bounds[1]))

            # Do not allow for connection choices, because the imputation mechanism allows for the effective selection
            # of other than the fixed value because it is not aware of the fixed nature of the variable
            # If this were to be supported again in the future, uncomment code related to _filter_conn_des_vectors_fixed
            for _, _, _, i_dv_start, i_dv_end, _ in self._conn_choice_data_map.values():
                if i_dv_start <= idx < i_dv_end:
                    raise RuntimeError('Design variable fixing not support for connection choices!')

            self._fixed_values[idx] = value

        self._update_comb_fixed_mask()
        clear_func_cache(self)

    def _update_comb_fixed_mask(self):
        fixed_choices = {}
        for i_dv, i_dec in enumerate(self._sel_choice_idx_map):
            if i_dv not in self._fixed_values:
                continue
            fixed_idx = int(self._fixed_values[i_dv])
            fixed_choices[i_dec] = fixed_idx

        self._comb_fixed_mask = self._hierarchy_analyzer.get_available_combinations_mask(fixed_choices)

    def free_des_var(self, des_var: DesVar):
        """Unfix a design variable"""
        self.fix_des_var(des_var, None)

    def is_fixed(self, des_var: DesVar) -> bool:
        """Returns whether a design variable is fixed"""
        idx = self.all_des_vars.index(des_var)
        return idx in self._fixed_values

    def fixed_value(self, des_var: DesVar) -> Union[float, int]:
        """Get the value of a fixed design variable"""
        idx = self.all_des_vars.index(des_var)
        if idx not in self._fixed_values:
            raise RuntimeError('Design variable not fixed: %r' % des_var)
        return self._fixed_values[idx]

    @staticmethod
    def _get_inactive_value(des_var: DesVar) -> Union[float, int]:
        return X_INACTIVE_IMPUTE if des_var.is_discrete else (sum(des_var.bounds)/2)

    @cached_property
    def connection_choice_nodes(self) -> List[ConnectionChoiceNode]:
        return self.graph.ordered_choice_nodes(
            [node for node in self.graph.choice_nodes if isinstance(node, ConnectionChoiceNode)])

    @cached_property
    def _choice_nodes(self) -> List[ChoiceNode]:
        choice_nodes: List[ChoiceNode] = []
        choice_nodes += self.selection_choice_nodes
        choice_nodes += self.connection_choice_nodes
        return choice_nodes

    @cached_property
    def design_variable_nodes(self) -> List[DesignVariableNode]:
        return self.graph.ordered_choice_nodes(self.graph.des_var_nodes)

    @cached_property
    def metric_nodes(self) -> List[MetricNode]:
        return sorted(self.graph.get_nodes_by_type(MetricNode), key=lambda n: n.name)

    def _get_metrics(self) -> List[Tuple[MetricNode, MetricType]]:
        permanent_nodes = self.permanent_nodes
        metrics = []
        for metric_node in self.metric_nodes:
            is_none = isinstance(metric_node.type, MetricType) and metric_node.type == MetricType.NONE
            if is_none:
                metric_type = MetricType.NONE
            else:
                obj = MetricType.OBJECTIVE if self._can_be_objective(metric_node, permanent_nodes) else MetricType.NONE
                constr = MetricType.CONSTRAINT if self._can_be_constraint(metric_node) else MetricType.NONE

                metric_type = obj | constr

            # If both is possible, check if there is a preference
            if metric_type == MetricType.OBJ_OR_CON and isinstance(metric_node.type, MetricType):
                metric_type = metric_node.type

            metrics.append((metric_node, metric_type))
        return metrics

    def _get_permanent_nodes(self):
        confirmed_initial_graph = self.graph.get_confirmed_graph()
        return set(confirmed_initial_graph.graph.nodes)

    @staticmethod
    def _can_be_objective(metric_node, permanent_nodes):
        """A metric can be an objective if it is part of the permanent nodes (i.e. is always there)."""
        return metric_node.dir is not None and metric_node in permanent_nodes

    @staticmethod
    def _can_be_constraint(metric_node):
        """A metric can be a constraint if a reference value has been defined."""
        return metric_node.dir is not None and metric_node.ref is not None

    def _categorize_metrics(self):
        objectives = []
        constraints = []
        for metric_node, metric_type in self._metrics:
            if metric_type & MetricType.OBJECTIVE:
                objective = Objective.from_metric_node(metric_node)

                # Need to choose?
                if metric_type & MetricType.CONSTRAINT:
                    constraint = Constraint.from_metric_node(metric_node)

                    item = self._choose_metric_type(objective, constraint)
                    if isinstance(item, Objective):
                        objectives.append(item)
                    else:
                        constraints.append(item)

                else:
                    objectives.append(objective)

            elif metric_type & MetricType.CONSTRAINT:
                constraints.append(Constraint.from_metric_node(metric_node))

        return objectives, constraints

    @cached_function
    def get_statistics(self):
        """Computes the following statistics:
        - Number of valid combinatorial design points (1)
        - Number of combinatorial design points (2)
        - Imputation ratio (2) / (1) --> lower is better
        - Information index (0 = 1 variable with n options; 1 = m variables with 2 options) --> higher is better
        - Distance correlation (-1 to 1, 1 means perfect correlation) --> higher is better
        - Number of existence patterns

        for: selection-choices, each connection-choice, additional design variables, total"""
        import pandas as pd

        stats = {'type': [], 'n_valid': [], 'n_declared': [], 'n_discrete': [], 'n_dim_cont': [], 'n_dim_cont_mean': [],
                 'n_exist': [], 'imp_ratio': [], 'imp_ratio_comb': [], 'imp_ratio_cont': [], 'inf_idx': [],
                 'dist_corr': [], 'encoder': []}

        # Selection-choices
        an = self._hierarchy_analyzer
        stats['type'].append('option-decisions')
        stats['n_valid'].append(an.n_combinations)
        stats['n_declared'].append(an.n_design_space)
        stats['n_discrete'].append(len(an.selection_choice_nodes))
        stats['n_dim_cont'].append(0)
        stats['n_dim_cont_mean'].append(0.)
        stats['imp_ratio'].append(an.imputation_ratio)
        stats['imp_ratio_comb'].append(an.imputation_ratio)
        stats['imp_ratio_cont'].append(1.)
        stats['inf_idx'].append(an.information_index)
        stats['dist_corr'].append(0.)
        stats['n_exist'].append(1)
        sel_choice_enc = self.encoder_type.value
        stats['encoder'].append(sel_choice_enc)

        # Permutation-options
        for i, connection_choice in enumerate(self.connection_choice_nodes):
            assignment_manager: AssignmentManagerBase
            assignment_manager, _, _, _, _, _ = self._conn_choice_data_map[connection_choice]
            stats['type'].append(f'permutation-decision-{i} {connection_choice.decision_id}')
            imp_ratio = assignment_manager.encoder.get_imputation_ratio()
            if np.isinf(imp_ratio):
                imp_ratio = 1e6
            stats['imp_ratio'].append(imp_ratio)
            stats['imp_ratio_comb'].append(imp_ratio)
            n_declared = assignment_manager.encoder.get_n_design_points()
            stats['n_declared'].append(n_declared)
            stats['n_discrete'].append(len(assignment_manager.design_vars))
            n_exist = len(list(assignment_manager.matrix_gen.iter_existence()))
            stats['n_exist'].append(n_exist)
            stats['n_valid'].append(int((n_declared*n_exist)/imp_ratio))
            stats['inf_idx'].append(assignment_manager.encoder.get_information_index())
            stats['dist_corr'].append(assignment_manager.encoder.get_distance_correlation())
            stats['encoder'].append(str(assignment_manager.encoder))

            stats['n_dim_cont'].append(0)
            stats['n_dim_cont_mean'].append(0.)
            stats['imp_ratio_cont'].append(1.)

        # Additional design variables
        n_dim_cont, n_dim_cont_mean, imp_ratio_cont = 0, 0., 1.
        if len(self.design_variable_nodes) > 0:
            n_comb_additional_dv, n_declared, n_discrete, n_dim_cont, n_dim_cont_mean, imp_ratio_cont = \
                self.get_additional_dv_stats()
            n_valid = int(np.sum(n_comb_additional_dv))
            if n_valid == 0 or n_discrete == 0:
                imp_ratio_comb = 1.
            else:
                imp_ratio_comb = n_declared*len(n_comb_additional_dv)/n_valid
            imp_ratio = imp_ratio_comb*imp_ratio_cont

            stats['type'].append('additional-dvs')
            stats['n_valid'].append(n_valid)
            stats['n_declared'].append(n_declared)
            stats['n_discrete'].append(n_discrete)
            stats['n_dim_cont'].append(n_dim_cont)
            stats['n_dim_cont_mean'].append(n_dim_cont_mean)
            stats['n_exist'].append(len(n_comb_additional_dv))
            stats['imp_ratio'].append(imp_ratio)
            stats['imp_ratio_comb'].append(imp_ratio_comb)
            stats['imp_ratio_cont'].append(imp_ratio_cont)
            stats['inf_idx'].append(1.)
            stats['dist_corr'].append(0.)
            stats['encoder'].append('')

        # Complete problem
        stats['type'].append('total-design-space')
        stats['n_valid'].append(self.get_n_valid_designs())
        stats['n_declared'].append(self.get_n_design_space())
        stats['n_discrete'].append(len(self.all_des_vars)-n_dim_cont)
        stats['n_dim_cont'].append(n_dim_cont)
        stats['n_dim_cont_mean'].append(n_dim_cont_mean)
        stats['imp_ratio'].append(self.get_imputation_ratio())
        stats['imp_ratio_comb'].append(self.get_imputation_ratio(include_cont=False))
        stats['imp_ratio_cont'].append(imp_ratio_cont)
        stats['inf_idx'].append(self.get_information_index())
        stats['dist_corr'].append(0.)
        stats['n_exist'].append(1)
        stats['encoder'].append(sel_choice_enc)

        # Problem with fixed design variables
        _, _, _, n_dim_cont_fix, n_dim_cont_mean_fix, imp_ratio_cont_fix = self.get_additional_dv_stats(with_fixed=True)
        stats['type'].append('total-design-problem')
        stats['n_valid'].append(self.get_n_valid_designs(with_fixed=True))
        stats['n_declared'].append(self.get_n_design_space(with_fixed=True))
        stats['n_discrete'].append(len(self.des_vars)-n_dim_cont_fix)
        stats['n_dim_cont'].append(n_dim_cont_fix)
        stats['n_dim_cont_mean'].append(n_dim_cont_mean_fix)
        stats['imp_ratio'].append(self.get_imputation_ratio(with_fixed=True))
        stats['imp_ratio_comb'].append(self.get_imputation_ratio(include_cont=False, with_fixed=True))
        stats['imp_ratio_cont'].append(imp_ratio_cont_fix)
        stats['inf_idx'].append(self.get_information_index(with_fixed=True))
        stats['dist_corr'].append(0.)
        stats['n_exist'].append(1)
        stats['encoder'].append(sel_choice_enc)

        df = pd.DataFrame(data=stats)
        df = df.set_index('type')
        return df

    def print_stats(self):
        """Get and print statistics"""
        import pandas as pd
        with pd.option_context('display.max_rows', None, 'display.max_columns', None,
                               'display.expand_frame_repr', False, 'max_colwidth', None):
            print(self.get_statistics())

    def _get_dv_n_opts(self, with_fixed=False, include_cont=False):
        des_vars = self.des_vars if with_fixed else self.all_des_vars
        if not include_cont:
            return [dv.n_opts for dv in des_vars if dv.is_discrete]
        return [dv.n_opts if dv.is_discrete else 2 for dv in des_vars]

    def get_n_design_space(self, with_fixed=False, include_cont=False) -> int:
        """Get the declared design space size (Cartesian product of discrete variables)"""
        n_opts = self._get_dv_n_opts(with_fixed=with_fixed, include_cont=include_cont)
        if len(n_opts) == 0:
            return 1
        return int(np.prod(n_opts, dtype=float))

    def get_information_index(self, with_fixed=False) -> float:
        return Encoder.calc_information_index(self._get_dv_n_opts(with_fixed=with_fixed))

    @cached_function
    @catch_memory_overflow
    def get_n_valid_designs(self, with_fixed=False, include_cont=False) -> int:
        """Get the number of valid (discrete) architectures"""
        # Possible combinations of selection-choices
        cutoff_mode = False
        n_combs = self._hierarchy_analyzer.n_combinations
        if self._memory_save_mode or n_combs > self._n_combs_cutoff:
            cutoff_mode = True
            n_combinations = np.ones((1,), dtype=float)*n_combs
        else:
            n_combinations = np.ones((n_combs,), dtype=float)
        if len(n_combinations) == 0:
            return 0

        # Possible combinations of independent connection-choices, per selection-choice combination
        for connection_choice in self.connection_choice_nodes:
            assignment_manager: AssignmentManagerBase
            assignment_manager, _, exist_map, i_dv_start, i_dv_end, _ = self._conn_choice_data_map[connection_choice]

            # If no accurate node existence information is available, assume the most combinations possible
            if isinstance(exist_map, dict) or cutoff_mode:
                n_comb_max = assignment_manager.matrix_gen.count_all_matrices(max_by_existence=True)
                n_combinations *= n_comb_max

            else:
                matrix_map = assignment_manager.matrix_gen.get_agg_matrix(cache=True)
                # dv_map = assignment_manager.get_all_design_vectors() if with_fixed else {}
                for i_comb, i_exist in enumerate(exist_map):
                    if i_exist == -1:  # Infeasible existence scheme
                        n_combinations[i_comb] = 0
                    else:
                        existence = assignment_manager.matrix_gen.existence_patterns.patterns[i_exist]
                        # if with_fixed:
                        #     des_vectors = self._filter_conn_des_vectors_fixed(
                        #         dv_map[existence], i_dv_start, i_dv_end, self._fixed_values)
                        #     n_dvs = des_vectors.shape[0]
                        # else:
                        n_dvs = matrix_map[existence].shape[0]
                        n_combinations[i_comb] *= n_dvs

        # Possible combinations of additional design variable values, per selection-choice combination
        if len(self.design_variable_nodes) > 0:
            n_comb_additional_dv, _, _, _, _, _ = self.get_additional_dv_stats(
                with_fixed=with_fixed, cont_as_discrete=include_cont)
            n_combinations *= n_comb_additional_dv

        # Determine which selection-choice combinations we should count
        if with_fixed and self._comb_fixed_mask is not None:
            n_combinations = n_combinations[self._comb_fixed_mask]

        return int(np.sum(n_combinations))

    # @staticmethod
    # def _filter_conn_des_vectors_fixed(des_vectors: np.ndarray, i_dv_start, i_dv_end, fixed_values: dict):
    #     for i, i_dv in enumerate(range(i_dv_start, i_dv_end)):
    #         fixed_value = fixed_values.get(i_dv)
    #         if fixed_value is not None:
    #             matched_mask = des_vectors[:, i] == fixed_value
    #
    #             # If the fixed value is zero, also match the inactive value (-1), because that is represented as zero
    #             if fixed_value == 0:
    #                 matched_mask |= des_vectors[:, i] == -1
    #
    #             des_vectors = des_vectors[matched_mask, :]
    #     return des_vectors

    @cached_function
    def get_all_discrete_x(self, with_fixed=True) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        """Returns all possible discrete design vectors. Warning: contains no limits on time or size!"""
        fixed_values = self._fixed_values if with_fixed else {}
        dv_idx_map = self.all_des_var_idx_map
        node_map = self.all_des_var_node_map
        n_dv = len(self.all_des_vars)

        # Selection choice design vectors
        choice_opt_idx = self._hierarchy_analyzer.get_choice_option_indices()
        if choice_opt_idx is None:
            return
        dv_sel = choice_opt_idx[:, ~self._sel_choice_is_forced]
        x = -np.ones((dv_sel.shape[0], n_dv), dtype=int)
        x[:, :dv_sel.shape[1]] = dv_sel[:, :]
        i_combs = np.arange(x.shape[0])

        # Filter fixed and infeasible values
        x_keep = self._existence_mask if with_fixed else self._existence_infeasibility_mask
        x = x[x_keep, :]
        i_combs = i_combs[x_keep]

        def _expand_at(i_dv_expand, x_expand, i_comb_expand):
            nonlocal x, i_combs
            n_repeat = np.ones((x.shape[0],), dtype=int)
            is_expanded = np.zeros((x.shape[0],), dtype=bool)
            n_expand = x_expand.shape[0]
            for i_comb_active in i_comb_expand:
                i_comb_act_mask = i_combs == i_comb_active
                n_repeat[i_comb_act_mask] = n_expand
                is_expanded[i_comb_act_mask] = True

            if ~np.any(is_expanded):
                return
            x = np.repeat(x, n_repeat, axis=0)
            i_combs = np.repeat(i_combs, n_repeat)

            x_columns = x[:, i_dv_expand]
            cumulative_idx = np.cumsum(n_repeat)-n_repeat
            for i_set, cum_idx in enumerate(cumulative_idx):
                if is_expanded[i_set]:
                    x_columns[cum_idx:cum_idx+n_expand, :] = x_expand

            x[:, i_dv_expand] = x_columns

        # Connection choices
        conn_choice_data_map = self._conn_choice_data_map
        conn_choice_existence = self._hierarchy_analyzer.get_nodes_existence(self.connection_choice_nodes)
        if conn_choice_existence is None:
            return
        for i_node, choice_node in enumerate(self.connection_choice_nodes):
            i_is_active = conn_choice_existence[:, i_node]

            assignment_manager: AssignmentManagerBase
            assignment_manager, conn_node_map, exist_map, i_dv_start, i_dv_end, all_conn_nodes = \
                conn_choice_data_map[choice_node]

            x_conn_map = assignment_manager.get_all_design_vectors()
            existence_patterns = assignment_manager.matrix_gen.existence_patterns.patterns

            if isinstance(exist_map, dict):
                raise RuntimeError('Unexpected existence map type')
            for i_exist in np.unique(exist_map):
                if i_exist == -1:
                    continue
                i_comb_exist = np.where((exist_map == i_exist) & i_is_active)[0]
                if len(i_comb_exist) == 0:
                    continue

                existence = existence_patterns[i_exist]
                x_existence = x_conn_map[existence]
                # x_existence = self._filter_conn_des_vectors_fixed(x_existence, i_dv_start, i_dv_end, fixed_values)

                _expand_at(np.arange(i_dv_start, i_dv_end), x_existence, i_comb_exist)

        # Additional discrete design variables
        dv_node_existence = self._hierarchy_analyzer.get_nodes_existence(self.design_variable_nodes)
        if dv_node_existence is None:
            return
        for i_dv, des_var_node in enumerate(self.design_variable_nodes):
            i_is_active = np.where(dv_node_existence[:, i_dv])[0]
            des_var = node_map[des_var_node][0]
            dv_idx = dv_idx_map[des_var]

            if not des_var.is_discrete:
                for i_comb in i_is_active:
                    x[i_combs == i_comb, dv_idx] = 0
                continue

            values = np.array([np.arange(des_var.n_opts)]).T
            if dv_idx in fixed_values:
                values = values[[fixed_values[dv_idx]], :]
            _expand_at([dv_idx], values, i_is_active)

        # Get activeness
        is_active = x != X_INACTIVE_VALUE  # Applies both for the hierarchy analyzer and assignment encoders
        x = x.astype(float)
        for i_dv, dv in enumerate(self.all_des_vars):
            inactive_value = self._get_inactive_value(dv)
            x[x[:, i_dv] == X_INACTIVE_VALUE, i_dv] = inactive_value

        # Return only non-fixed design variables
        is_free = np.array([i for i in range(n_dv) if i not in fixed_values], dtype=int)
        x = x[:, is_free]
        is_active = is_active[:, is_free]
        return x, is_active

    @cached_function
    @catch_memory_overflow
    def get_additional_dv_stats(self, with_fixed=False, cont_as_discrete=False):
        cutoff_mode = False
        n_comb = self._hierarchy_analyzer.n_combinations
        multiplier = 1
        if self._memory_save_mode or n_comb > self._n_combs_cutoff:
            multiplier = n_comb
            n_comb = 1
            cutoff_mode = True

        if len(self.design_variable_nodes) == 0:
            nodes_existence = np.zeros((n_comb, 0), dtype=bool)
        else:
            try:
                nodes_existence = self._hierarchy_analyzer.get_nodes_existence(self.design_variable_nodes) \
                    if not cutoff_mode else None
            except MemoryError:
                nodes_existence = None
                multiplier = n_comb
                n_comb = 1

            # If precise information is unavailable, assume all nodes are active
            if nodes_existence is None:
                nodes_existence = np.ones((n_comb, len(self.design_variable_nodes)), dtype=bool)*multiplier

        cont_mask = np.array([not node.is_discrete for node in self.design_variable_nodes], dtype=bool)
        n_opts = []
        n_comb_additional_dv = np.ones(nodes_existence.shape, dtype=int)
        for i_dv_node, dv_node in enumerate(self.design_variable_nodes):
            # Do not count if fixed
            i_dv = self.all_des_vars.index(self.all_des_var_node_map[dv_node][0])
            if with_fixed and i_dv in self._fixed_values:
                n_opts.append(1)
                continue

            if not cont_mask[i_dv_node] or cont_as_discrete:
                n_combs_dv = 2 if cont_mask[i_dv_node] else len(dv_node.options)
                n_comb_additional_dv[nodes_existence[:, i_dv_node], i_dv_node] = n_combs_dv
                n_opts.append(n_combs_dv)
            else:
                n_comb_additional_dv[~nodes_existence[:, i_dv_node], i_dv_node] = 0

        n_comb_discrete = n_comb_additional_dv if cont_as_discrete else n_comb_additional_dv[:, ~cont_mask]
        n_discrete = n_comb_discrete.shape[1]
        if n_comb_discrete.shape[1] == 0:
            n_comb_additional_dv_cum = np.ones((nodes_existence.shape[0],), dtype=int)
            n_declared = 0
        else:
            n_comb_additional_dv_cum = np.prod(n_comb_discrete, axis=1, dtype=float)
            n_declared = int(np.prod(n_opts, dtype=float)) if len(n_opts) > 0 else 1

        n_comb_cont = n_comb_additional_dv[:, cont_mask]
        if n_comb_cont.shape[1] == 0:
            n_dim_cont, n_dim_cont_mean, imp_ratio_cont = 0, 0., 1.
        else:
            n_dim_cont = n_comb_cont.shape[1]
            n_dim_cont_mean = np.sum(n_comb_cont)/n_comb_cont.shape[0]
            imp_ratio_cont = n_dim_cont/n_dim_cont_mean if n_dim_cont_mean > 0 else 1.

        return n_comb_additional_dv_cum, n_declared, n_discrete, n_dim_cont, n_dim_cont_mean, imp_ratio_cont

    def get_imputation_ratio(self, with_fixed=False, include_cont=True) -> float:
        """
        Get the imputation ratio: the ratio between the declared and valid design space sizes.
        A value of 1 indicates there is no design space hierarchy, values greater than 1 indicate hierarchy.
        """
        n_valid = self.get_n_valid_designs(with_fixed=with_fixed)
        if n_valid == 0:
            return 1.
        imp_ratio = self.get_n_design_space(with_fixed=with_fixed)/n_valid

        if include_cont:
            _, _, _, _, _, imp_ratio_cont = self.get_additional_dv_stats(with_fixed=with_fixed)
            imp_ratio *= imp_ratio_cont

        return imp_ratio

    def get_random_design_vector(self) -> Sequence[Union[int, float]]:
        """Generate a random design vector"""
        return [dv.rand() for dv in self.des_vars]

    def _get_all_des_var_values(self, des_var_values: Sequence[Union[int, float]]) -> List[Union[int, float]]:
        fixed_values = self._fixed_values
        i_value = 0
        values = []
        for i, des_var in enumerate(self.all_des_vars):
            if i in fixed_values:
                values.append(fixed_values[i])
            else:
                values.append(des_var_values[i_value])
                i_value += 1
        return values

    def get_graph(self, des_var_values: Sequence[Union[int, float]], create=True) \
            -> Tuple[DSGType, List[Union[int, float]], List[bool]]:
        """Creates a DSG instance from a design vector; optionally only returning the imputed design vector and
        activeness vector without actually creating the graph"""

        if len(self.des_vars) != len(des_var_values):
            raise ValueError('Incorrect number of design variable values: %d != %d' %
                             (len(des_var_values), len(self.des_vars)))

        des_var_values_in = des_var_values
        des_var_values = self._get_all_des_var_values(des_var_values)
        des_vars = self.all_des_vars
        if len(des_vars) != len(des_var_values):
            raise RuntimeError

        # Get graph for the selection-choice nodes
        sel_choice_opt_idx = [0 for _ in range(len(self.selection_choice_nodes))]
        is_fixed = [False for _ in range(len(sel_choice_opt_idx))]
        for i_dv, i_dec in enumerate(self._sel_choice_idx_map):
            sel_choice_opt_idx[i_dec] = int(des_var_values[i_dv])
            is_fixed[i_dec] = i_dv in self._fixed_values

        sel_choice_is_active = graph_instance = i_comb = None
        try:
            if not create:
                graph_instance = None
                sel_choice_opt_idx, sel_choice_is_active, i_comb = self._hierarchy_analyzer.get_opt_idx(
                    sel_choice_opt_idx, mask=self._existence_mask, is_fixed=is_fixed, exclude=self._excluded_cache)

                # If the combination idx is not known (fast encoder),
                # we need to actually create the graph to know which nodes exist
                if i_comb is None:
                    create = True

            if create:
                graph_instance, sel_choice_opt_idx, sel_choice_is_active, i_comb = self._hierarchy_analyzer.get_graph(
                    sel_choice_opt_idx, mask=self._existence_mask, is_fixed=is_fixed, exclude=self._excluded_cache)

        except RuntimeError:
            print(f'Error occurred while getting graph for DV: {des_var_values}')
            raise

        opt_dec_used_values: List[Optional[int]] = \
            [int(val) for val in list(np.array(sel_choice_opt_idx)[self._sel_choice_idx_map])]
        for i_dv, i_dec in enumerate(self._sel_choice_idx_map):
            if not sel_choice_is_active[i_dec]:
                opt_dec_used_values[i_dv] = None

        opt_dec_existence_key = tuple(sel_choice_opt_idx)

        # Associate rest of design variable to choice nodes
        idx_map = self.all_des_var_idx_map
        node_map = self.all_des_var_node_map

        # Track actually used design variable values
        used_values: List[Optional[Union[int, float]]] = [None for _ in range(len(des_vars))]
        used_values[:len(opt_dec_used_values)] = opt_dec_used_values

        # Make connection-choices
        graph_cache = self._graph_cache
        prev_values = opt_dec_used_values
        conn_choice_data_map = self._conn_choice_data_map

        if i_comb is not None:
            conn_choice_existence = \
                self._hierarchy_analyzer.get_nodes_existence(self.connection_choice_nodes, i_comb=i_comb)
        elif graph_instance is not None:
            conn_choice_existence = [choice_node in graph_instance.choice_nodes
                                     for choice_node in self.connection_choice_nodes]
        else:
            raise RuntimeError('Either combination idx or graph instance is needed!')

        for i_conn_choice, choice_node in enumerate(self.connection_choice_nodes):
            # Check if choice still exists
            if not conn_choice_existence[i_conn_choice]:
                continue
            if graph_instance is not None and choice_node not in graph_instance.choice_nodes:
                raise RuntimeError('Connection choice not does not exist!')

            assignment_manager: AssignmentManagerBase
            assignment_manager, conn_node_map, exist_map, i_dv_start, i_dv_end, all_conn_nodes = \
                conn_choice_data_map[choice_node]

            if isinstance(exist_map, dict):
                if graph_instance is None:
                    raise RuntimeError('Graph instance expected!')
                existing_nodes = graph_instance.graph.nodes
                nodes_exist_mask = tuple(node in existing_nodes for node in all_conn_nodes)
                i_exist_pattern = exist_map[nodes_exist_mask]
            else:
                i_exist_pattern = exist_map[i_comb]

            # If the pattern has no solutions, either try to find another graph or raise error if this shouldn't happen
            if i_exist_pattern == -1:
                if i_comb is None:
                    self._excluded_cache.add(tuple(sel_choice_opt_idx))
                    return self.get_graph(des_var_values_in, create=create)
                else:
                    raise RuntimeError('Infeasible graph specified!')

            choice_des_vector = np.array(des_var_values[i_dv_start:i_dv_end]).astype(int)
            existence = assignment_manager.matrix_gen.existence_patterns.patterns[i_exist_pattern]
            choice_des_vector, choice_is_active, conn_edges = \
                assignment_manager.get_conn_idx(choice_des_vector, existence=existence)

            node_edges = [(conn_node_map[0][conn_edge[0]], conn_node_map[1][conn_edge[1]]) for conn_edge in conn_edges]

            if graph_instance is not None:
                cache_key = (choice_node, opt_dec_existence_key, tuple(choice_des_vector), tuple(prev_values))
                if cache_key in graph_cache:
                    graph_instance = graph_cache[cache_key].copy()
                else:
                    graph_cache[cache_key] = graph_instance = \
                        graph_instance.get_for_apply_connection_choice(choice_node, node_edges, validate=False)

            used_values[i_dv_start:i_dv_end] = [
                int(val) if choice_is_active[i_dv] else None for i_dv, val in enumerate(choice_des_vector)]
            prev_values = prev_values+list(choice_des_vector)

        # Set values of additional design variables
        if i_comb is not None:
            dv_node_existence = self._hierarchy_analyzer.get_nodes_existence(self.design_variable_nodes, i_comb=i_comb)
        elif graph_instance is not None:
            dv_node_existence = [dv_node in graph_instance.des_var_nodes for dv_node in self.design_variable_nodes]
        else:
            raise RuntimeError('Either combination idx or graph instance is needed!')

        if np.any(dv_node_existence):
            if graph_instance is not None:
                graph_instance = graph_instance.copy()
            for i_dv, des_var_node in enumerate(self.design_variable_nodes):
                if not dv_node_existence[i_dv]:
                    continue
                if graph_instance is not None and des_var_node not in graph_instance.des_var_nodes:
                    raise RuntimeError('Des var node not found!')
                if des_var_node not in node_map:
                    raise ValueError('Unexpected design variable node: %s' % des_var_node.str_context())
                des_var = node_map[des_var_node][0]
                dv_idx = idx_map[des_var]
                des_var_value = des_var_values[dv_idx]

                des_var_value = int(des_var_value) if des_var.is_discrete else float(des_var_value)
                if graph_instance is not None:
                    graph_instance.set_des_var_value(des_var_node, des_var_value)
                    used_values[dv_idx] = graph_instance.des_var_value(des_var_node)
                else:
                    used_values[dv_idx], _ = des_var_node.correct_value(des_var_value)

        # Set all unused design variables to their inactive values
        is_active = [used_value is not None for used_value in used_values]
        for i, used_value in enumerate(used_values):
            if used_value is None:
                used_values[i] = self._get_inactive_value(des_vars[i])

        # Get non-fixed used values
        used_values = [value for i, value in enumerate(used_values) if i not in self._fixed_values]
        is_active = [is_act for i, is_act in enumerate(is_active) if i not in self._fixed_values]

        return graph_instance, used_values, is_active

    def _encode_connection_choice(self, connection_choice_node: ConnectionChoiceNode):

        # Get all nodes involved in the connection and their existence specifications
        settings, node_map, existence_map, all_conn_nodes = connection_choice_node.get_assignment_encoding_args(
            self.graph, hierarchy_analyzer=self._hierarchy_analyzer)

        # Get assignment manager by automatically selecting the best encoder
        selector = EncoderSelector(settings)
        if self._reset_conn_encoder_cache:
            selector.reset_cache()
        assignment_manager = selector.get_best_assignment_manager()

        # Remove any existence pattern that results in an infeasible graph (i.e. no connections are possible)
        agg_matrix = assignment_manager.matrix_gen.get_agg_matrix(cache=True)
        for i_exist, existence in enumerate(assignment_manager.matrix_gen.existence_patterns.patterns):
            if agg_matrix[existence].shape[0] == 0:
                if isinstance(existence_map, dict):
                    existence_map = {exist_mask: -1 if i_pattern == i_exist else i_pattern
                                     for exist_mask, i_pattern in existence_map.items()}
                else:
                    existence_map[existence_map == i_exist] = -1

        # Convert to design variables
        base_name = connection_choice_node.decision_id
        des_vars = []
        for i, dv in enumerate(assignment_manager.design_vars):
            options = list(range(dv.n_opts))
            name = f'{base_name}_{i}'
            des_vars.append(DesVar.from_choice_node(
                connection_choice_node, options=options, name=name, conditionally_active=dv.conditionally_active))

        return assignment_manager, des_vars, node_map, existence_map, all_conn_nodes

    def _choose_metric_type(self, objective: Objective, constraint: Constraint) -> Union[Objective, Constraint]:
        raise RuntimeError('Cannot choose whether to interpret metric as objective or as constraint: '
                           '%r' % objective.name)



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import *
import networkx as nx
from collections import defaultdict
from adsg_core.graph.traversal import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *
from adsg_core.graph.incompatibility import *

__all__ = ['get_mod_apply_selection_choice', 'get_mod_apply_connection_choice', 'get_mod_apply_choice_constraint',
           'NoOptionError']


def get_mod_apply_choice_constraint(graph: nx.MultiDiGraph, start_nodes: set, choice_node: ChoiceNode,
                                    removed_option_nodes: List[DSGNode]):
    removed_edges = set()
    removed_nodes = set()

    # Process decision-option constraints
    for edge in iter_out_edges(graph, choice_node):
        # Check if this target node is removed
        if edge[1] not in removed_option_nodes:
            continue

        derived_edges, derived_nodes = get_derived_edges_for_edge(
            graph, edge, start_nodes, removed_edges=removed_edges, removed_nodes=removed_nodes)
        removed_edges |= derived_edges
        removed_nodes |= derived_nodes

    return removed_edges, removed_nodes


class NoOptionError(ValueError):
    pass


def get_mod_apply_selection_choice(
        graph: nx.MultiDiGraph, start_nodes: set, choice_node: ChoiceNode, target_option_node: DSGNode = None,
        choice_con_map: List[Tuple[SelectionChoiceNode, List[DSGNode]]] = None, only_added=False) -> tuple:

    # Check outgoing edges
    choice_out_edges = set(iter_out_edges(graph, choice_node))
    option_nodes = {edge[1] for edge in choice_out_edges}

    # If there are no options, remove the decision node and mark the originating node as infeasible
    if len(option_nodes) == 0:
        removed_nodes = {choice_node}

        originating_nodes = list(graph.predecessors(choice_node))
        added_edges = {get_edge_for_type(
            list(start_nodes)[0], originating_node, EdgeType.INCOMPATIBILITY, choice_node=choice_node)
            for originating_node in originating_nodes}

        return set(), removed_nodes, added_edges

    if target_option_node not in option_nodes:
        raise NoOptionError(f'Node ({target_option_node!s}) is not an option of choice node: '
                            f'{choice_node!s} -> {option_nodes!s}')

    removed_edges = set()
    removed_nodes = set()
    in_edges = list(iter_in_edges(graph, choice_node))
    added_edges = {get_edge(in_edge[0], target_option_node) for in_edge in in_edges}

    if only_added:
        return set(), set(), added_edges

    # Process decision-option constraints
    if choice_con_map is not None:
        for constrained_dec_node, removed_options in choice_con_map:
            if constrained_dec_node not in graph.nodes:
                continue
            for constrained_out_edge in iter_out_edges(graph, constrained_dec_node):

                if constrained_out_edge[1] in removed_options:
                    choice_out_edges.add(constrained_out_edge)

    # Remove derived nodes
    for edge in choice_out_edges:
        # If this is the target edge, do not remove the derived nodes
        if edge[0] == choice_node and edge[1] == target_option_node:
            continue

        derived_edges, derived_nodes = get_derived_edges_for_edge(
            graph, edge, start_nodes, removed_edges=removed_edges, removed_nodes=removed_nodes)
        removed_edges |= derived_edges
        removed_nodes |= derived_nodes

    removed_nodes.add(choice_node)

    # Process incompatibility constraints
    confirmed_start_nodes = start_nodes | {target_option_node}
    try:
        removed_nodes |= get_mod_nodes_remove_incompatibilities(graph, confirmed_start_nodes, removed_edges)
    except IncompatibilityError as e:
        removed_nodes |= e.removed_nodes
        added_edges |= e.edges

    return removed_edges, removed_nodes, added_edges


def get_mod_apply_connection_choice(graph: nx.MultiDiGraph, choice_node: ConnectionChoiceNode,
                                    edges: Sequence[Tuple[ConnectorNode, ConnectorNode]]) -> tuple:

    in_nodes = {edge[0] for edge in iter_in_edges(graph, choice_node)}
    out_nodes = {edge[1] for edge in iter_out_edges(graph, choice_node)}

    for edge in edges:
        if edge[0] not in in_nodes or (edge[1] is not None and edge[1] not in out_nodes):
            raise ValueError('Node not part of connection choice')

    removed_nodes = {choice_node}

    # Create edges with correct keys
    added_edges = set()
    edge_key = defaultdict(int)
    for edge in edges:
        if edge[1] is not None:
            added_edges.add(get_edge(edge[0], edge[1], key=edge_key[edge], is_conn=True))
            edge_key[edge] += 1

    # Remove exclusion edges
    removed_edges = set(choice_node.get_excluded_edges(graph)) | set(choice_node.get_deriving_edges(graph))

    return removed_edges, removed_nodes, added_edges



import numpy as np
from typing import *
from collections import defaultdict
from adsg_core.optimization.assign_enc.lazy_encoding import *
from adsg_core.optimization.assign_enc.lazy.encodings.conn_idx import ConnCombsEncoder
from adsg_core.optimization.assign_enc.eager.encodings.grouped_base import GroupedEncoder

__all__ = ['LazyAmountFirstEncoder', 'FlatLazyAmountEncoder', 'TotalLazyAmountEncoder', 'SourceLazyAmountEncoder',
           'SourceTargetLazyAmountEncoder', 'FlatLazyConnectionEncoder']


NList = List[Tuple[Tuple[int, ...], Tuple[int, ...]]]


class LazyAmountEncoder:
    """Base class for encoding a set of src and tgt connection amounts"""

    def _get_n_src_tgt_array(self, n_src_n_tgt: NList) -> np.ndarray:
        return np.array([n_src_conn+n_tgt_conn for n_src_conn, n_tgt_conn in n_src_n_tgt], dtype=int)

    def encode_prepare(self):
        pass

    def encode(self, n_src, n_tgt, n_src_n_tgt: NList, existence: NodeExistence, n_declared_start=None) \
            -> List[DiscreteDV]:
        raise NotImplementedError

    def decode(self, vector: DesignVector, existence: NodeExistence) \
            -> Optional[Tuple[DesignVector, Tuple[int, ...], Tuple[int, ...]]]:
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class LazyConnectionEncoder:
    """Base class for encoding connection selection"""

    def encode_prepare(self):
        pass

    def encode(self, n_src_n_tgt: NList, n_matrix_map: Dict[tuple, np.ndarray], existence: NodeExistence,
               encoder: QuasiLazyEncoder) -> List[DiscreteDV]:
        raise NotImplementedError

    def decode(self, n_src_conn, n_tgt_conn, matrices: np.ndarray, vector: DesignVector, encoder: QuasiLazyEncoder,
               existence: NodeExistence) -> Optional[Tuple[DesignVector, np.ndarray]]:
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class LazyAmountFirstEncoder(QuasiLazyEncoder):
    """Base class for an encoder that first defines one or more design variables for selecting the amount of
    connections, and then selects the connection pattern. Generates matrices only when needed (on demand)."""

    def __init__(self, imputer: LazyImputer, amount_encoder: LazyAmountEncoder, conn_encoder: LazyConnectionEncoder):
        super().__init__(imputer)
        self.amount_encoder = amount_encoder
        self.conn_encoder = conn_encoder
        self._i_dv_amount = {}
        self._n_dv_amount_expand = {}
        self._n_dv_amount = {}
        self._i_dv_conn = {}
        self._n_dv_conn_expand = {}
        self._n_matrix_map = {}

    def _encode_prepare(self):
        super()._encode_prepare()

        self._i_dv_amount = {}
        self._n_dv_amount_expand = {}
        self._n_dv_amount = {}
        self._i_dv_conn = {}
        self._n_dv_conn_expand = {}
        self._n_matrix_map = {}

        self.amount_encoder.encode_prepare()
        self.conn_encoder.encode_prepare()

    def _encode_matrix(self, matrix: np.ndarray, existence: NodeExistence) -> List[DiscreteDV]:
        mat_n_src, mat_n_tgt = np.sum(matrix, axis=2), np.sum(matrix, axis=1)

        n_src_n_tgt_map = defaultdict(list)
        for i, n_src in enumerate(mat_n_src):
            n_src_n_tgt_map[(tuple(n_src), tuple(mat_n_tgt[i]))].append(i)
        n_src_n_tgt = list(n_src_n_tgt_map.keys())

        dv_amount = self.amount_encoder.encode(self.n_src, self.tgt, n_src_n_tgt, existence)
        dv_amount, self._i_dv_amount[existence], self._n_dv_amount_expand[existence] = self._filter_dvs(dv_amount)
        self._n_dv_amount[existence] = len(dv_amount)

        self._n_matrix_map[existence] = n_matrix_map = \
            {n_conn_key: matrix[indices, :, :] for n_conn_key, indices in n_src_n_tgt_map.items()}
        dv_conn = self.conn_encoder.encode(n_src_n_tgt, n_matrix_map, existence, self)
        dv_conn, self._i_dv_conn[existence], self._n_dv_conn_expand[existence] = self._filter_dvs(dv_conn)

        return dv_amount+dv_conn

    def _decode_matrix(self, vector: DesignVector, matrix: np.ndarray, existence: NodeExistence) \
            -> Optional[Tuple[DesignVector, np.ndarray]]:
        if existence not in self._n_dv_amount:
            return
        n_dv_amt = self._n_dv_amount[existence]

        # Decode connection amounts
        amount_vector_expanded = \
            self._unfilter_dvs(vector[:n_dv_amt], self._i_dv_amount[existence], self._n_dv_amount_expand[existence])

        n_src_tgt_conn = self.amount_encoder.decode(amount_vector_expanded, existence)
        if n_src_tgt_conn is None:
            return
        imp_amt_vector_expanded, n_src_conn, n_tgt_conn = n_src_tgt_conn
        imp_amount_vector_sel = imp_amt_vector_expanded[self._i_dv_amount[existence]]
        imp_amount_vector = list(imp_amount_vector_sel)+[X_INACTIVE_VALUE]*(n_dv_amt-len(imp_amount_vector_sel))

        n_dv_conn = len(vector)-n_dv_amt
        conn_vector_expanded = self._unfilter_dvs(
            vector[n_dv_amt:], self._i_dv_conn[existence], self._n_dv_conn_expand[existence])

        # Decode connections matrix
        n_conn_key = (tuple(n_src_conn), tuple(n_tgt_conn))
        matrices = self._n_matrix_map.get(existence, {}).get(n_conn_key)
        if matrices is None:
            return

        matrix_data = self.conn_encoder.decode(n_src_conn, n_tgt_conn, matrices, conn_vector_expanded, self, existence)
        if matrix_data is None:
            return
        imp_conn_vector_expanded, matrix = matrix_data
        imp_conn_vector_sel = imp_conn_vector_expanded[self._i_dv_conn[existence]]
        imp_conn_vector = list(imp_conn_vector_sel)+[X_INACTIVE_VALUE]*(n_dv_conn-len(imp_conn_vector_sel))

        imp_vector = imp_amount_vector+imp_conn_vector
        return imp_vector, matrix

    @staticmethod
    def group_by_values(values: np.ndarray, n_declared_start=None) -> np.ndarray:
        return GroupedEncoder.group_by_values(values, n_declared_start=n_declared_start)

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r}, {self.amount_encoder!r}, {self.conn_encoder!r})'

    def __str__(self):
        return f'Lazy Amount First + {self._imputer!s}: {self.amount_encoder!s}; {self.conn_encoder!s}'


class FlatLazyAmountEncoder(LazyAmountEncoder):

    def __init__(self):
        self._n_src_n_tgt = {}

    def encode_prepare(self):
        self._n_src_n_tgt = {}

    def encode(self, n_src, n_tgt, n_src_n_tgt: NList, existence: NodeExistence, n_declared_start=None) \
            -> List[DiscreteDV]:
        self._n_src_n_tgt[existence] = n_src_n_tgt
        return [DiscreteDV(n_opts=len(n_src_n_tgt), conditionally_active=False)]

    def decode(self, vector: DesignVector, existence: NodeExistence) \
            -> Optional[Tuple[DesignVector, Tuple[int, ...], Tuple[int, ...]]]:
        n_existence = self._n_src_n_tgt[existence]
        if len(n_existence) == 0:
            return

        idx = vector[0]
        if idx >= len(n_existence):
            idx = len(n_existence)-1
            vector = np.array([idx])

        n_src, n_tgt = n_existence[idx]
        return vector, n_src, n_tgt

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Flat Amount'


class GroupedLazyAmountEncoder(LazyAmountEncoder):

    def __init__(self):
        self._dv_val_map = {}

    def encode_prepare(self):
        self._dv_val_map = {}

    def encode(self, n_src, n_tgt, n_src_n_tgt: NList, existence: NodeExistence, n_declared_start=None) \
            -> List[DiscreteDV]:
        n_src_tgt_arr = self._get_n_src_tgt_array(n_src_n_tgt)
        if len(n_src_n_tgt) == 0:
            return []

        dv_group_values = self._get_dv_group_values(n_src, n_tgt, n_src_tgt_arr)
        dv_values = LazyAmountFirstEncoder.group_by_values(dv_group_values, n_declared_start=n_declared_start)
        self._dv_val_map[existence] = (ConnCombsEncoder.get_dv_map_for_lookup(dv_values), n_src_n_tgt)

        cond_active = np.any(dv_values == X_INACTIVE_VALUE, axis=0)
        return [DiscreteDV(n_opts=n, conditionally_active=cond_active[i])
                for i, n in enumerate(np.max(dv_values, axis=0)+1)]

    def decode(self, vector: DesignVector, existence: NodeExistence) \
            -> Optional[Tuple[DesignVector, Tuple[int, ...], Tuple[int, ...]]]:

        if existence not in self._dv_val_map:
            return
        dv_map, n_src_n_tgt = self._dv_val_map[existence]
        idx_and_dv = ConnCombsEncoder.lookup_dv(dv_map, np.array(vector))
        if idx_and_dv is not None:
            i, dv = idx_and_dv
            n_src, n_tgt = n_src_n_tgt[i]
            return dv, n_src, n_tgt

    def _get_dv_group_values(self, n_src, n_tgt, n_src_tgt_arr: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        raise NotImplementedError


class TotalLazyAmountEncoder(GroupedLazyAmountEncoder):

    def _get_dv_group_values(self, n_src, n_tgt, n_src_tgt_arr: np.ndarray) -> np.ndarray:
        return np.column_stack([
            np.sum(n_src_tgt_arr[:, :n_src], axis=1),
            np.arange(n_src_tgt_arr.shape[0]),
        ])

    def __str__(self):
        return 'Total Amount'


class SourceLazyAmountEncoder(GroupedLazyAmountEncoder):

    def _get_dv_group_values(self, n_src, n_tgt, n_src_tgt_arr: np.ndarray) -> np.ndarray:
        return np.column_stack([
            n_src_tgt_arr[:, :n_src],
            np.arange(n_src_tgt_arr.shape[0]),
        ])

    def __str__(self):
        return 'Source Amount'


class SourceTargetLazyAmountEncoder(GroupedLazyAmountEncoder):

    def _get_dv_group_values(self, n_src, n_tgt, n_src_tgt_arr: np.ndarray) -> np.ndarray:
        return n_src_tgt_arr

    def __str__(self):
        return 'Source Target Amount'


class FlatLazyConnectionEncoder(LazyConnectionEncoder):

    def __init__(self):
        self._n_exist_max = {}
        super().__init__()

    def encode_prepare(self):
        self._n_exist_max = {}

    def encode(self, n_src_n_tgt: NList, n_matrix_map: Dict[tuple, np.ndarray], existence: NodeExistence,
               encoder: QuasiLazyEncoder) -> List[DiscreteDV]:

        if len(n_matrix_map) == 0:
            n_matrix_max = 1
        else:
            n_matrix_max = max([matrix.shape[0] for matrix in n_matrix_map.values()])

        self._n_exist_max[existence] = n_matrix_max
        return [DiscreteDV(n_opts=n_matrix_max, conditionally_active=False)] if n_matrix_max > 1 else []

    def decode(self, n_src_conn, n_tgt_conn, matrices: np.ndarray, vector: DesignVector, encoder: QuasiLazyEncoder,
               existence: NodeExistence) -> Optional[Tuple[DesignVector, np.ndarray]]:

        if matrices.shape[0] == 0:
            return vector, np.zeros((len(n_src_conn), len(n_tgt_conn)), dtype=int)
        if len(vector) == 0:
            return vector, matrices[0, :, :]

        if matrices.shape[0] > self._n_exist_max.get(existence, 0):
            raise RuntimeError(f'Sub-matrix found ({n_src_conn}, {n_tgt_conn}) with '
                               f'{matrices.shape[0]} > {self._n_exist_max.get(existence, 0)} possibilities')

        i_conn = vector[0]
        if i_conn >= matrices.shape[0]:
            i_conn = matrices.shape[0]-1
            vector = np.array([i_conn])

        return vector, matrices[i_conn, :, :]

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Flat Conn'



from adsg_core.graph.adsg_nodes import *

__all__ = ['SupNode']


class SupNode(NamedNode):
    """
    Basic node in a Supplementary DSG.

    Reference can be anything that helps the user in linking/associating the node to whatever it represents, as long as
    it is serializable/hashable and has a repr() representation.
    SupNodes with the same repr(ref) and name will be recognized as the same node.
    """

    def __init__(self, name: str, ref=None, **kwargs):
        self.ref = ref
        obj_id = self._get_obj_id(name, ref)
        super().__init__(name, obj_id=obj_id, **kwargs)

    def _get_obj_id(self, name, ref):
        return f'{name}|{ref!r}'

    def get_export_color(self) -> str:
        return '#C8E6C9'  # Green 100



import itertools
import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.lazy_encoding import *

__all__ = ['LazyFirstImputer']


class LazyFirstImputer(LazyImputer):
    """Imputer that chooses the first possible matrix."""

    def _impute(self, vector: DesignVector, matrix: Optional[np.ndarray], existence: NodeExistence,
                validate: Callable[[np.ndarray], bool], tried_vectors: Set[Tuple[int, ...]]) \
            -> Tuple[DesignVector, np.ndarray]:

        if len(vector) == 0:
            return vector, np.zeros((0, 0), dtype=int)

        cache_key = ('found_first', hash(existence))
        if cache_key in self._impute_cache:
            return self._impute_cache[cache_key]

        # Loop through possible design vectors until one is found that has not been tried yet
        for dv in itertools.product(*[list(range(dv.n_opts)) for dv in self._get_des_vars(existence)[::-1]]):
            # Validate this design vector and associated matrix
            vector = np.array(dv[::-1])
            vector, matrix = self._decode(vector, existence)
            if validate(matrix):
                self._impute_cache[cache_key] = vector, matrix
                return vector, matrix

        self._impute_cache[cache_key] = vector, np.zeros((0, 0), dtype=int)
        return self._impute_cache[cache_key]

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'First Imp'



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.lazy_encoding import *

__all__ = ['EnumOrdinalEncoder']


class EnumOrdinalEncoder(QuasiLazyEncoder):
    """Defines one design variable with the same number of options as the number of possible assignment patterns."""

    def _encode_matrix(self, matrix: np.ndarray, existence: NodeExistence) -> List[DiscreteDV]:
        n_mat = matrix.shape[0]
        if n_mat <= 1:
            return []
        return [DiscreteDV(n_opts=n_mat, conditionally_active=False)]

    def _decode_matrix(self, vector: DesignVector, matrix: np.ndarray, existence: NodeExistence) \
            -> Optional[Tuple[DesignVector, np.ndarray]]:
        i_mat = vector[0] if len(vector) > 0 else 0
        if i_mat >= matrix.shape[0]:
            return [matrix.shape[0]-1], matrix[-1, :, :]
        return vector, matrix[i_mat, :, :]

    def _do_generate_random_dv_mat(self, n: int, existence: NodeExistence, matrix: np.ndarray,
                                   design_vars: List[DiscreteDV]) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        if n < matrix.shape[0]:
            i_selected = np.random.choice(matrix.shape[0], n, replace=False)
        else:
            i_selected = np.arange(matrix.shape[0])

        design_vectors = np.array([i_selected]).T
        matrices = matrix[i_selected, :, :]
        return design_vectors, matrices

    def _do_get_all_design_vectors(self, existence: NodeExistence, matrix: np.ndarray, design_vars: List[DiscreteDV]) \
            -> np.ndarray:
        design_vectors = np.array([np.arange(matrix.shape[0])]).T
        return design_vectors

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r})'

    def __str__(self):
        return f'Enum Ord + {self._imputer!s}'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import math
from typing import *
from adsg_core.graph.adsg import DSGType
from adsg_core.graph.adsg_nodes import MetricNode
from adsg_core.optimization.dv_output_defs import *
from adsg_core.optimization.graph_processor import *

__all__ = ['DSGEvaluator', 'ADSGEvaluator']


class DSGEvaluator(GraphProcessor):
    """
    Base class for implementing an evaluator that directly evaluates DSG instances.
    Override _evaluate to implement the evaluation.

    Extends `GraphProcessor`, so all its functions are also available.
    """

    def get_problem(self, n_parallel=None, parallel_processes=True):
        """Get an SBArchOpt problem instance."""
        from adsg_core.optimization.problem import DSGArchOptProblem
        return DSGArchOptProblem(self, n_parallel=n_parallel, parallel_processes=parallel_processes)

    def _choose_metric_type(self, objective: Objective, constraint: Constraint) -> Union[Objective, Constraint]:
        raise RuntimeError(f'Metric {objective.name} can either be an objective or a constraint! '
                           f'Specify the metric type using node.type = MetricType.x')

    def evaluate(self, dsg: DSGType) -> Tuple[List[float], List[float]]:
        """
        Evaluate a DSG instance. Returns a list of objective values and a list of constraint values.
        """

        # Evaluate the DSG instance
        metric_nodes = dsg.metric_nodes
        value_map = self._evaluate(dsg, metric_nodes)

        # Set metric values
        for metric_node in metric_nodes:
            dsg.set_metric_value(metric_node, value_map.get(metric_node, math.nan))

        # Associate values to objectives
        objective_values = [value_map.get(objective.node, math.nan) for objective in self.objectives]

        # Associate values to constraints: if the constraint does not exist in this DSG instance, set value to ref
        constraint_values = [value_map.get(constraint.node, math.nan)
                             if constraint.node in metric_nodes else constraint.ref
                             for constraint in self.constraints]

        return objective_values, constraint_values

    def _evaluate(self, dsg: DSGType, metric_nodes: List[MetricNode]) -> Dict[MetricNode, float]:
        """
        Implement this function to provide DSG evaluation.
        Should return a mapping from metric node to float (NaN is allowed).
        """
        raise NotImplementedError


ADSGEvaluator = DSGEvaluator  # Backward compatibility



import itertools
import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.lazy_encoding import *

__all__ = ['LazyDeltaImputer']


class LazyDeltaImputer(LazyImputer):
    """Imputes by looking for the first valid design vector created by iterating over delta values."""

    _n_max_tries_default = 10000

    def __init__(self, n_max_tries=None):
        self.n_max_tries = n_max_tries if n_max_tries is not None else self._n_max_tries_default
        super().__init__()

    def _impute(self, vector: DesignVector, matrix: Optional[np.ndarray], existence: NodeExistence,
                validate: Callable[[np.ndarray], bool], tried_vectors: Set[Tuple[int, ...]]) -> Tuple[DesignVector, np.ndarray]:

        # Check if we have any design variables
        if len(vector) == 0:
            return vector, np.zeros((0, 0), dtype=int)

        # Determine delta values to try out
        def _sort_by_dist(des_var_delta):
            i_sorted = np.argsort(np.abs(des_var_delta))
            return des_var_delta[i_sorted]

        delta_values = [_sort_by_dist(np.arange(dv.n_opts)-vector[i])
                        for i, dv in enumerate(self._get_des_vars(existence))]

        # Loop over all delta values
        n_tries, n_tries_max = 0, self.n_max_tries
        for dv_delta in self.yield_dv_delta_product(delta_values):
            dv = vector+np.array(dv_delta)

            # Validate associated matrix
            results = self._decode(dv, existence)
            if results is not None:
                dv, matrix = results
                if validate(matrix):
                    return dv, matrix

            # Limit the amount of tries
            n_tries += 1
            if n_tries > n_tries_max:
                invalid_matrix = -np.ones((len(self._matrix_gen.src), len(self._matrix_gen.tgt)), dtype=int)
                return vector, invalid_matrix

        return vector, np.zeros((0, 0), dtype=int)

    def yield_dv_delta_product(self, delta_values):
        first = True
        for dv_delta in itertools.product(*delta_values):
            if first:  # Skip first as this one has a delta of 0
                first = False
                continue
            yield dv_delta

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Delta Imp'



import copy
import numba
import warnings
import contextlib
import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.matrix import *
from dataclasses import dataclass
from scipy.spatial import distance
from scipy.stats import pearsonr, ConstantInputWarning

warnings.filterwarnings('ignore', category=ConstantInputWarning)

__all__ = ['DiscreteDV', 'DesignVector', 'PartialDesignVector', 'MatrixSelectMask', 'X_INACTIVE_VALUE',
           'IsActiveVector', 'EagerImputer', 'Encoder', 'EagerEncoder', 'filter_design_vectors', 'flatten_matrix',
           'NodeExistence', 'DetectedHighImpRatio']


@dataclass
class DiscreteDV:
    n_opts: int
    conditionally_active: bool

    def get_random(self):
        return np.random.randint(0, self.n_opts)


DesignVector = Union[List[int], np.ndarray]
IsActiveVector = Union[List[bool], np.ndarray]
PartialDesignVector = List[Optional[int]]
MatrixSelectMask = np.ndarray

X_INACTIVE_VALUE = -1


def filter_design_vectors(design_vectors: np.ndarray, vector: PartialDesignVector) -> MatrixSelectMask:
    int_vector = np.array([-1 if val is None else val for val in vector], dtype=int)
    return _filter_design_vectors(design_vectors, int_vector)


@numba.njit()
def _filter_design_vectors(design_vectors: np.ndarray, vector: np.ndarray) -> MatrixSelectMask:
    """Filter matrices along the first dimension given a design vector. Returns a mask of selected matrices."""

    dv_mask = np.ones((design_vectors.shape[0],), dtype=numba.types.bool_)
    for i, value in enumerate(vector):
        if value != -1:
            # Select design vectors that have the targeted value for this design variable
            dv_mask[dv_mask] = np.bitwise_and(dv_mask[dv_mask], design_vectors[dv_mask, i] == value)
    return dv_mask


class EagerImputer:
    """Base class for imputing design vectors to select existing matrices."""

    def __init__(self):
        self._matrix: Optional[Dict[NodeExistence, np.ndarray]] = None
        self._design_vectors: Optional[Dict[NodeExistence, np.ndarray]] = None
        self._design_vectors_zero: Optional[Dict[NodeExistence, np.ndarray]] = None
        self._design_vars: Optional[List[DiscreteDV]] = None

    def initialize(self, matrix: Dict[NodeExistence, np.ndarray], design_vectors: Dict[NodeExistence, np.ndarray],
                   design_vectors_zero: Dict[NodeExistence, np.ndarray], design_vars: List[DiscreteDV]):
        self._matrix = matrix
        self._design_vectors = design_vectors
        self._design_vectors_zero = design_vectors_zero
        self._design_vars = design_vars

    def _get_design_vectors(self, existence: NodeExistence, zero=True) -> np.ndarray:
        dv_map = self._design_vectors_zero if zero else self._design_vectors
        if existence not in dv_map:
            return np.empty((0, 0), dtype=int)
        return dv_map[existence]

    def _get_matrix(self, existence: NodeExistence) -> np.ndarray:
        return EagerEncoder.get_matrix_for_existence(self._matrix, existence)[0]

    def _filter_design_vectors(self, vector: PartialDesignVector, existence: NodeExistence) -> MatrixSelectMask:
        design_vectors = self._get_design_vectors(existence)
        vector = vector[:design_vectors.shape[1]]
        return filter_design_vectors(design_vectors, vector)

    def _return_imputation(self, i_dv: int, existence: NodeExistence) -> Tuple[DesignVector, np.ndarray]:
        design_vectors = self._get_design_vectors(existence, zero=False)
        matrix = self._get_matrix(existence)
        return design_vectors[i_dv, :], matrix[i_dv, :, :]

    def impute(self, vector: DesignVector, existence: NodeExistence, matrix_mask: MatrixSelectMask) \
            -> Tuple[DesignVector, np.ndarray]:
        """Return a new design vector and associated assignment matrix (n_src x n_tgt)"""
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class DetectedHighImpRatio(RuntimeError):
    """Exception thrown when an encoder detects a high imputation ratio during the encoding process"""

    def __init__(self, encoder: Type['Encoder'], imp_ratio: float, msg: str = None):
        self.encoder = encoder
        self.imp_ratio = imp_ratio
        self.msg = msg
        super().__init__()

    def __str__(self):
        msg_str = f': {self.msg}' if self.msg is not None else ''
        return f'{self.__class__.__name__} when encoding "{self.encoder!s}" ({self.imp_ratio}){msg_str}'

    def __repr__(self):
        return f'{self.__class__.__name__}({self.encoder!r}, imp_ratio={self.imp_ratio}, msg={self.msg})'


class Encoder:

    # Set the imputation ratio limit for early detection (only if the encoder supports it)
    # If detected, a DetectedHighImpRatio exception is thrown
    # Use `with_early_detect_high_imp_ratio` as a context manager
    _early_detect_high_imp_ratio = None

    @classmethod
    @contextlib.contextmanager
    def with_early_detect_high_imp_ratio(cls, high_imp_ratio):
        initial_value = cls._early_detect_high_imp_ratio
        cls._early_detect_high_imp_ratio = high_imp_ratio
        try:
            yield high_imp_ratio
        finally:
            cls._early_detect_high_imp_ratio = initial_value

    @property
    def design_vars(self) -> List[DiscreteDV]:
        raise NotImplementedError

    def get_for_imputer(self, imputer):
        raise NotImplementedError

    def get_random_design_vector(self) -> DesignVector:
        return [dv.get_random() for dv in self.design_vars]

    def get_n_design_points(self) -> int:
        return self.calc_n_declared_design_points(self.design_vars)

    @staticmethod
    def calc_n_declared_design_points(design_vars: List[DiscreteDV]) -> int:
        if len(design_vars) == 0:
            return 1
        return int(np.prod([dv.n_opts for dv in design_vars], dtype=float))

    def get_information_index(self) -> float:
        return self.calc_information_index([dv.n_opts for dv in self.design_vars])

    @staticmethod
    def calc_information_index(n_opts: List[int]) -> float:
        n_dv = len(n_opts)
        if n_dv == 0:
            return 1.
        n_combinations = np.prod(n_opts, dtype=float)
        if n_combinations <= 2:
            return 1.
        n_dv_max = np.log2(n_combinations)
        return (n_dv-1.)/(n_dv_max-1.)

    def get_distance_correlation(self, n=100, minimum=False, n_max=2000, n_samples_min=150, limit=.005) -> Optional[float]:
        """Correlation between distances between design vectors and associated matrices;
        optionally returning the minimum across all existence schemes"""
        dv_dist_all = None
        mat_dist_all = None
        n_check = max(2, int(np.floor(n_samples_min/n)))
        sampled_dvs = {}
        n_iter_max = int(np.ceil(n_max / n))
        metric_values, i_iter = None, 0
        n_has_zero = 0
        for i_iter in range(n_iter_max):
            # Extend current samples
            dv_dist, mat_dist, all_zero = self._get_distance_correlation(n=n, sampled_dvs=sampled_dvs)
            if dv_dist_all is None:
                dv_dist_all = dv_dist
                mat_dist_all = mat_dist
                metric_values = np.nan*np.empty((n_iter_max, len(dv_dist) if minimum else 1))
            else:
                dv_dist_all = [dists+dv_dist[i_mat] for i_mat, dists in enumerate(dv_dist_all)]
                mat_dist_all = [dists+mat_dist[i_mat] for i_mat, dists in enumerate(mat_dist_all)]

            # Calculate new metric value
            if minimum:
                for i_mat, dv_dists in enumerate(dv_dist_all):
                    metric_values[i_iter, i_mat] = self._calc_distance_corr(dv_dists, mat_dist_all[i_mat])
            else:
                metric_values[i_iter, 0] = self._calc_distance_corr(dv_dist_all, mat_dist_all)

            # Track the nr of times we have no new design vectors sampled
            if all_zero:
                n_has_zero += 1
                if n_has_zero >= 2:
                    break

            # Check convergence
            if i_iter+1 >= n_check:
                n_samples = [len(samples) for samples in dv_dist_all]
                if (min(n_samples) if minimum else sum(n_samples)) >= n_samples_min:
                    last_values = metric_values[:i_iter+1, :][-n_check:, :]
                    max_diff = np.max(np.max(last_values, axis=0)-np.min(last_values, axis=0))
                    if max_diff <= limit:
                        break

                # Correlation of 1 is usually quickly found
                if np.max(np.abs(metric_values-1)) < 1e-3:
                    break

        metrics_mean = np.mean(metric_values[:i_iter+1, :][-n_check:, :], axis=0)
        return float(np.min(metrics_mean) if minimum else metrics_mean[0])

    def _get_distance_correlation(self, n: int, sampled_dvs=None) -> Tuple[List[List[float]], List[List[float]], bool]:
        if sampled_dvs is None:
            sampled_dvs = {}
        dv_dist_all = []
        mat_dist_all = []
        all_zero = True
        for matrix, des_vectors in self._iter_sampled_dv_mat(n, sampled_dvs):
            if len(des_vectors) <= 1 or len(matrix) <= 1:
                dv_dist_all.append([])
                mat_dist_all.append([])
                continue

            all_zero = False
            des_vectors[des_vectors == X_INACTIVE_VALUE] = 0
            dv_dist_all.append(list(self._calc_internal_dv_distance(des_vectors)[np.triu_indices(des_vectors.shape[0], k=1)]))
            mat_dist_all.append(list(self._calc_internal_distance(matrix)[np.triu_indices(matrix.shape[0], k=1)]))

        return dv_dist_all, mat_dist_all, all_zero

    def _iter_sampled_dv_mat(self, n: int, sampled_dvs: dict):
        raise NotImplementedError

    @staticmethod
    def _calc_internal_dv_distance(arr: np.ndarray) -> np.ndarray:
        return distance.cdist(arr, arr, 'cityblock')  # Manhattan distance

    @staticmethod
    def _calc_internal_distance(arr: np.ndarray) -> np.ndarray:
        return distance.cdist(arr, arr, 'cityblock')  # Manhattan distance

    @staticmethod
    def _calc_distance_corr(dist1, dist2) -> float:
        if isinstance(dist1, list) and len(dist1) > 0 and isinstance(dist1[0], list):
            dist1 = np.array([val for values in dist1 for val in values])
            dist2 = np.array([val for values in dist2 for val in values])
        if len(dist1) != len(dist2):
            raise RuntimeError(f'Different lengths: {len(dist1)} != {len(dist2)}')
        if len(dist1) < 2:
            return 1.

        corr = pearsonr(dist1, dist2).statistic
        return 1. if np.isnan(corr) else corr

    @staticmethod
    def _pad_dv_map(dv_map: Dict[NodeExistence, np.ndarray], n_dv: int) -> Dict[NodeExistence, np.ndarray]:
        padded_dv_map = {}
        for existence, design_vectors in dv_map.items():
            n_extra = n_dv-design_vectors.shape[1]
            if n_extra > 0:
                if design_vectors.shape[0] == 0:
                    design_vectors = np.zeros((0, n_dv), dtype=int)
                else:
                    design_vectors = np.column_stack([
                        design_vectors, np.ones((design_vectors.shape[0], n_extra), dtype=int)*X_INACTIVE_VALUE])
            padded_dv_map[existence] = design_vectors
        return padded_dv_map

    def get_imputation_ratio(self, per_existence=False) -> float:
        """Ratio of the total design space size to the actual amount of possible connections"""
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class EagerEncoder(Encoder):
    """Base class that encodes assignment matrices to discrete design variables."""

    def __init__(self, imputer: EagerImputer, matrix: np.ndarray = None):
        self._matrix = {}
        self._design_vectors = {}
        self._design_vectors_zeros = {}
        self._design_vector_map = {}
        self._design_vars = []
        self._imputer = imputer

        if matrix is not None:
            self.matrix = matrix

    @property
    def matrix(self) -> Dict[NodeExistence, np.ndarray]:
        return self._matrix

    @matrix.setter
    def matrix(self, matrix: MatrixMapOptional):
        if isinstance(matrix, np.ndarray):
            matrix = {NodeExistence(): matrix}

        # Encode separately for each existence mode
        self._matrix = matrix
        self._design_vectors = des_vectors = {existence: self._encode(mat) for existence, mat in matrix.items()}

        self._design_vectors_zeros = des_vectors_zero = {}
        for existence, des_vec in des_vectors.items():
            des_vec_zeros = des_vec.copy()
            des_vec_zeros[des_vec_zeros == X_INACTIVE_VALUE] = 0
            des_vectors_zero[existence] = des_vec_zeros

        # Inactive variables in the design vectors are encoded as -1 (X_INACTIVE_VALUE), however, for lookup
        # (_design_vector_map) they are stored as 0's, as that is what real design vectors would also have
        self._design_vector_map = {existence: {tuple(dv.tolist()): i for i, dv in enumerate(des_vec)}
                                   for existence, des_vec in des_vectors_zero.items()}

        self._design_vars = self.get_design_variables(des_vectors)
        self._imputer.initialize(matrix, self._design_vectors, self._design_vectors_zeros, self._design_vars)

    @property
    def padded_design_vectors(self) -> Dict[NodeExistence, np.ndarray]:
        return self._pad_dv_map(self._design_vectors, len(self.design_vars))

    def set_imputer(self, imputer: EagerImputer):
        self._imputer = imputer
        if self._matrix is not None:
            self._imputer.initialize(self._matrix, self._design_vectors, self._design_vectors_zeros, self._design_vars)

    def get_for_imputer(self, imputer: EagerImputer):
        encoder: EagerEncoder = copy.deepcopy(self)
        encoder.set_imputer(imputer)
        return encoder

    @property
    def n_mat_max(self) -> int:
        return max([matrix.shape[0] for matrix in self._matrix.values()])

    @property
    def design_vars(self) -> List[DiscreteDV]:
        return self._design_vars

    def _iter_sampled_dv_mat(self, n: int, sampled_dvs: dict):
        for existence, matrix in self._matrix.items():
            matrix = self.flatten_matrix(matrix)
            des_vectors = self._design_vectors_zeros[existence]
            if des_vectors.shape[0] == 0:
                continue

            if existence not in sampled_dvs:
                sampled_dvs[existence] = np.ones((matrix.shape[0],), dtype=bool)
            unsampled_mask = sampled_dvs[existence]

            matrix = matrix[unsampled_mask, :]
            des_vectors = des_vectors[unsampled_mask, :]

            if matrix.shape[0] > n:
                i_random = np.random.choice(matrix.shape[0], size=n, replace=False)
                matrix = matrix[i_random, :]
                des_vectors = des_vectors[i_random, :]
                unsampled_mask[np.where(unsampled_mask)[0][i_random]] = False

            yield matrix, des_vectors

    def get_imputation_ratio(self, per_existence=False) -> float:
        n_design_points = self.get_n_design_points()
        n_total = []
        n_valid = []
        for matrix in self._matrix.values():
            n_total.append(n_design_points)
            n_valid.append(matrix.shape[0])
        if per_existence:
            return min([n_tot/n_valid[i] if n_valid[i] > 0 else np.inf for i, n_tot in enumerate(n_total)])
        return sum(n_total)/sum(n_valid) if sum(n_valid) > 0 else np.inf

    def _correct_vector_size(self, vector: DesignVector) -> Tuple[DesignVector, int, int]:
        n_dv = len(self.design_vars)
        vector, n_extra = self.correct_vector_size(n_dv, vector)
        return vector, n_dv, n_extra

    @staticmethod
    def correct_vector_size(n_dv, vector: DesignVector) -> Tuple[DesignVector, int]:
        n_extra = len(vector)-n_dv
        vector = vector[:n_dv]
        return vector, n_extra

    @staticmethod
    def correct_vector_bounds(vector: DesignVector, design_vars: List[DiscreteDV]) -> Tuple[DesignVector, bool]:
        correct_vector = vector.copy()
        is_corrected = False
        for i, dv in enumerate(design_vars):
            if correct_vector[i] < 0:
                correct_vector[i] = 0
                is_corrected = True

            elif correct_vector[i] >= dv.n_opts:
                correct_vector[i] = dv.n_opts-1
                is_corrected = True

        return correct_vector, is_corrected

    def get_matrix(self, vector: DesignVector, existence: NodeExistence = None,
                   matrix_mask: MatrixSelectMask = None) -> Tuple[DesignVector, np.ndarray]:
        """Select a connection matrix (n_src x n_tgt) and impute the design vector if needed.
        Inactive design variables get a value of -1."""

        vector, n_dv, n_extra = self._correct_vector_size(vector)
        extra_vector = [X_INACTIVE_VALUE]*n_extra
        vector, _ = self.correct_vector_bounds(vector, self.design_vars)

        i_mat, existence = self.get_matrix_index(vector, existence=existence, matrix_mask=matrix_mask)
        matrix, existence = self._get_matrix_for_existence(existence)

        # If this existence mode has no matrices, return the zero vector
        if not self._has_existence(existence):
            vector = [X_INACTIVE_VALUE]*len(vector)
            null_matrix = np.zeros((matrix.shape[1], matrix.shape[2]), dtype=int)
            return vector+extra_vector, null_matrix

        # If no matrix can be found, impute
        if i_mat is None:
            if matrix_mask is None:
                matrix_mask = np.ones((matrix.shape[0],), dtype=bool)

            # If the mask filters out all design vectors, there is no need to try imputing
            elif np.all(~matrix_mask):
                null_matrix = matrix[0, :, :]*0
                return [X_INACTIVE_VALUE]*len(vector)+extra_vector, null_matrix

            vector, matrix = self._imputer.impute(vector, existence, matrix_mask)
            if len(vector) < n_dv:
                vector = list(vector) + [X_INACTIVE_VALUE]*(n_dv-len(vector))
            return list(vector)+extra_vector, matrix

        # Design vector directly maps to possible matrix
        vector = self._correct_vector(vector, existence)
        return list(vector)+extra_vector, matrix[i_mat, :, :]

    def is_valid_vector(self, vector: DesignVector, existence: NodeExistence = None,
                        matrix_mask: MatrixSelectMask = None) -> bool:

        vector, n_dv, n_extra = self._correct_vector_size(vector)
        vector, is_corrected = self.correct_vector_bounds(vector, self.design_vars)
        if is_corrected:
            return False

        i_mat, existence = self.get_matrix_index(vector, existence=existence, matrix_mask=matrix_mask)
        if i_mat is None:
            return False

        vector = list(self._design_vectors[existence][i_mat, :])
        corr_vector = self._correct_vector(vector, existence)
        if not np.all(corr_vector == vector):
            return False
        return True

    def get_matrix_index(self, vector: DesignVector, existence: NodeExistence = None,
                         matrix_mask: MatrixSelectMask = None) -> Tuple[Optional[int], NodeExistence]:
        matrix, existence = self._get_matrix_for_existence(existence)
        if matrix_mask is None:
            matrix_mask = np.ones((matrix.shape[0],), dtype=bool)
        i_mat, = np.where(matrix_mask & self._filter_full_design_vector(matrix, vector, existence=existence))
        if len(i_mat) > 1:
            raise RuntimeError(f'Design vector maps to more than one matrix: {vector}')

        # Only if we find exactly one corresponding matrix, it is indeed a valid design vector
        return i_mat[0] if len(i_mat) == 1 else None, existence

    def _has_existence(self, existence: NodeExistence = None):
        if existence is None:
            existence = NodeExistence()
        return existence in self._matrix

    def _get_matrix_for_existence(self, existence: NodeExistence = None) -> Tuple[np.ndarray, NodeExistence]:
        if len(self._matrix) == 0:
            raise RuntimeError('Matrix not set!')
        return self.get_matrix_for_existence(self._matrix, existence=existence)

    def _correct_vector(self, vector: DesignVector, existence: NodeExistence) -> DesignVector:
        """Set unused design variables to zero (can happen when we are in an existence mode with less design variables
        than the max)"""
        n_dv = 0
        if existence in self._design_vectors:
            n_dv = self._design_vectors[existence].shape[1]

        corrected_vector = vector.copy()
        for i_dv in range(n_dv, len(vector)):
            corrected_vector[i_dv] = X_INACTIVE_VALUE
        return corrected_vector

    @staticmethod
    def get_matrix_for_existence(matrix_map: MatrixMap, existence: NodeExistence = None)\
            -> Tuple[np.ndarray, NodeExistence]:
        if existence is None:
            existence = NodeExistence()
        if existence not in matrix_map:
            first_matrix = list(matrix_map.values())[0]
            return np.empty((0, first_matrix.shape[1], first_matrix.shape[2]), dtype=int), existence
        return matrix_map[existence], existence

    def _filter_full_design_vector(self, matrix: np.ndarray, vector: DesignVector,
                                   existence: NodeExistence = None) -> MatrixSelectMask:
        if existence is None:
            existence = NodeExistence()

        mask = np.zeros((matrix.shape[0],), dtype=bool)
        if existence not in self._design_vector_map:
            return mask
        design_vectors = self._design_vector_map[existence]
        n_dv = self._design_vectors[existence].shape[1]

        i_mat = None
        if n_dv == 0:
            if len(self._matrix[existence]) == 1:
                i_mat = 0
        else:
            i_mat = design_vectors.get(tuple(vector[:n_dv]))

        if i_mat is not None:
            mask[i_mat] = True
        return mask

    @classmethod
    def get_design_variables(cls, design_vectors: Dict[NodeExistence, np.ndarray]) -> List[DiscreteDV]:
        """Convert possible design vectors to design variable definitions"""
        design_vars_list = []
        for des_vectors in design_vectors.values():
            if des_vectors.shape[0] == 0 or des_vectors.shape[1] == 0:
                continue

            # Check if all design vectors are unique
            seen_dvs = set()
            for dv in des_vectors:
                dv_key = tuple(dv.tolist())
                if dv_key in seen_dvs:
                    raise RuntimeError('Not all design vectors are unique!')
                seen_dvs.add(dv_key)

            # Check bounds
            if np.min(des_vectors[des_vectors != X_INACTIVE_VALUE]) != 0:
                raise RuntimeError('Design variables should start at zero!')

            n_opts_max = np.max(des_vectors, axis=0)+1
            is_cond_act = np.any(des_vectors == X_INACTIVE_VALUE, axis=0)
            design_vars_list.append([DiscreteDV(n_opts=n_opts, conditionally_active=is_cond_act[i_dv])
                                     for i_dv, n_opts in enumerate(n_opts_max)])

        # Merge design variables
        design_vars = cls.merge_design_vars(design_vars_list)
        for dv in design_vars:
            if dv.n_opts <= 1:
                raise RuntimeError('All design variables must have at least two options')
        return design_vars

    @staticmethod
    def merge_design_vars(design_vars_list: List[List[DiscreteDV]]) -> List[DiscreteDV]:
        if len(design_vars_list) == 0:
            return []
        n_dv_max = max([len(dvs) for dvs in design_vars_list])
        n_opts = np.zeros((len(design_vars_list), n_dv_max), dtype=int)
        is_cond_act = np.ones(n_opts.shape, dtype=bool)
        for i, dvs in enumerate(design_vars_list):
            n_opts[i, :len(dvs)] = [dv.n_opts for dv in dvs]
            is_cond_act[i, :len(dvs)] = [dv.conditionally_active for dv in dvs]

        n_opts_max = np.max(n_opts, axis=0)
        is_cond_act = np.any(is_cond_act, axis=0)
        return [DiscreteDV(n_opts=n, conditionally_active=is_cond_act[i]) for i, n in enumerate(n_opts_max)]

    @staticmethod
    def flatten_matrix(matrix: np.ndarray) -> np.ndarray:
        return flatten_matrix(matrix)

    @staticmethod
    def normalize_design_vectors(design_vectors: np.ndarray, remove_gaps=True) -> np.ndarray:
        """Move lowest values to 0 and eliminate value gaps."""

        # Check if there are no design vectors defined
        if design_vectors.shape[0] == 0:
            return design_vectors

        # Move to zero
        design_vectors = design_vectors.copy()
        if not remove_gaps:
            for i_dv in range(design_vectors.shape[1]):
                is_act_mask = design_vectors[:, i_dv] != X_INACTIVE_VALUE
                min_value = np.min(design_vectors[is_act_mask, i_dv])
                design_vectors[is_act_mask, i_dv] -= min_value

        # Remove gaps (also moves to zero)
        else:
            for i_dv in range(design_vectors.shape[1]):
                des_var = design_vectors[:, i_dv].copy()
                unique_values = np.sort(np.unique(des_var[des_var != X_INACTIVE_VALUE]))
                for i_unique, value in enumerate(unique_values):
                    design_vectors[des_var == value, i_dv] = i_unique

        # Remove design variables with not enough options
        no_opts_mask = np.max(design_vectors, axis=0) == 0
        design_vectors = design_vectors[:, ~no_opts_mask]

        return design_vectors

    def _encode(self, matrix: np.ndarray) -> np.ndarray:
        """
        Encode a matrix of size n_patterns x n_src x n_tgt as discrete design variables.
        Returns the list of design vectors for each matrix in a n_patterns x n_dv array.
        Assumes values range between 0 and the number of options per design variable, or -1 for inactive variables.
        """
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


def flatten_matrix(matrix: np.ndarray) -> np.ndarray:
    """Helper function that flattens matrices in the higher dimensions: n_mat x n_src x n_tgt --> n_mat x n_src*n_tgt"""
    return matrix.reshape(matrix.shape[0], np.prod(matrix.shape[1:]))



"""
MIT License

Copyright: (c) 2023, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import pickle
import itertools
import numpy as np
from typing import *
from adsg_core.optimization.sel_choice_enc.encoder import *
from adsg_core.optimization.sel_choice_enc.brute_force import *

__all__ = ['export_results', 'assert_behavior', 'validate_results', 'create_from_results']


def assert_behavior(encoder: SelectionChoiceEncoder):
    n_dv = encoder.n_choices
    n_opts = encoder.design_variable_options
    assert len(n_opts) == n_dv

    is_forced = encoder.design_variable_is_forced
    assert len(is_forced) == n_dv
    if n_dv > 0:
        not_forced = ~np.array(is_forced, dtype=bool)
        assert np.any(not_forced)  # At least one DV should be non-forced

        assert np.min(np.array(n_opts)[not_forced]) >= 2

    n_valid = encoder.n_valid
    assert 0 <= n_valid <= encoder.n_design_space
    if n_dv == 0:
        assert n_valid <= 1

    n_test_all = 1000
    x_all = status_all = None
    node_ids = encoder.node_ids
    if n_valid <= n_test_all:
        x_all, status_all = encoder.all_design_vectors_and_statuses
        assert x_all.shape == (n_valid, n_dv)
        assert status_all.shape == (n_valid, len(node_ids))

        for n in range(1, 4):
            for i_nodes in itertools.combinations(range(len(node_ids)), n):
                node_ids_query = [node_ids[i_n] for i_n in i_nodes]
                existence = encoder.get_node_existence_mask(node_ids_query)
                assert np.all(existence == (status_all[:, i_nodes] == Diag.CONFIRMED.value))

                for i_dv in np.random.randint(0, n_valid, 10):
                    existence_i = encoder.get_node_existence_mask(node_ids_query, i_dv=i_dv)
                    assert np.all(existence_i == (status_all[i_dv, i_nodes] == Diag.CONFIRMED.value))

    i_dv_test = np.arange(n_valid) if n_valid <= n_test_all else np.random.randint(0, n_valid, n_test_all)
    for i_dv in i_dv_test:
        i_exclude_all = set(range(n_valid))-{i_dv}
        dv, i_dv_ = encoder.get_design_vector_index([0]*n_dv, i_dv_excluded=i_exclude_all)
        assert i_dv_ == i_dv
        assert len(dv) == n_dv

        if x_all is not None:
            assert np.all(x_all[i_dv, :] == dv)
            node_exists = encoder.get_node_existence_mask(list(node_ids), i_dv=i_dv)
            assert np.all(node_exists == (status_all[i_dv, :] == Diag.CONFIRMED.value))

        dv_imp = np.array(dv, dtype=int)
        dv_imp[dv_imp == X_INACTIVE_VALUE] = 0
        dv_imp = list(dv_imp)

        dv_idx_set = encoder.get_design_vector_indices({ic: io for ic, io in enumerate(dv_imp)})
        assert dv_idx_set == {i_dv}

        dv_, i_dv_ = encoder.get_design_vector_index(dv_imp)
        assert i_dv_ == i_dv
        assert np.all(np.array(dv_) == np.array(dv))

        # Check that forced design variables are ignored
        for ix in np.where(is_forced)[0]:
            for value in range(n_opts[ix]):
                dv_mod = list(dv_imp)
                dv_mod[ix] = value

                dv_, i_dv_ = encoder.get_design_vector_index(dv_mod)
                assert i_dv_ == i_dv
                assert np.all(np.array(dv_) == np.array(dv))


def export_results(encoder: SelectionChoiceEncoder):
    x_all, status_all = encoder.all_design_vectors_and_statuses
    return {
        'kwargs': {
            'influence_matrix': encoder.influence_matrix,
            'node_ids': encoder.node_ids,
            'choice_opt_map': encoder.choice_opt_map,
            'choice_constraint_map': encoder.choice_constraint_map,
        },
        'dv_options': encoder.design_variable_options,
        'is_forced': encoder.design_variable_is_forced,
        'n_valid': encoder.n_valid,
        'x_all': x_all,
        'status_all': status_all,
    }


def create_from_results(encoder_factory: Callable[[Any], SelectionChoiceEncoder], results):
    if not isinstance(results, dict):
        with open(results, 'rb') as fp:
            results = pickle.load(fp)

    return encoder_factory(**results['kwargs']), results


def validate_results(encoder_factory: Callable[[Any], SelectionChoiceEncoder], results):
    encoder, results = create_from_results(encoder_factory, results)

    assert np.all(np.array(encoder.design_variable_options) == np.array(results['dv_options']))
    assert np.all(np.array(encoder.design_variable_is_forced) == np.array(results['is_forced']))
    assert encoder.n_valid == results['n_valid']

    x_ref: np.ndarray = results['x_all']
    x_ref_map = {tuple(xi): ix for ix, xi in enumerate(x_ref)}
    status_ref: np.ndarray = results['status_all']

    # Ignore actual order
    x_all, status_all = encoder.all_design_vectors_and_statuses
    assert x_all.shape == x_ref.shape
    i_ref = np.zeros((x_ref.shape[0],), dtype=int)
    for ix, xi in enumerate(x_all):
        ix_ref = x_ref_map.get(tuple(xi))
        assert ix_ref is not None
        i_ref[ix] = ix_ref
    assert len(np.unique(i_ref)) == len(i_ref)

    x_ref = x_ref[i_ref, :]
    status_ref = status_ref[i_ref, :]
    assert np.all(x_all == x_ref)
    assert np.all(status_all == status_ref)

    for i_dv in range(x_all.shape[0]):
        dv = x_all[i_dv, :]
        dv_imp = dv.copy()
        dv_imp[dv_imp == X_INACTIVE_VALUE] = 0

        dv_, i_dv_ = encoder.get_design_vector_index(list(dv_imp))
        assert i_dv_ == i_dv
        assert np.all(np.array(dv_) == dv)

    n_opts = encoder.design_variable_options
    for n in range(1, 4):
        for i_x in itertools.combinations(range(x_all.shape[1]), n):
            for i_x_values in itertools.product(*[list(range(n_opts[i])) for i in i_x]):
                choice_opt_idx_map = {i_choice: i_x_values[i] for i, i_choice in enumerate(i_x)}
                i_dv_matched = BruteForceSelectionChoiceEncoder._match_dv_indices(x_all, choice_opt_idx_map)
                assert i_dv_matched == encoder.get_design_vector_indices(choice_opt_idx_map)

    node_ids = encoder.node_ids
    for n in range(1, 4):
        for i_nodes in itertools.combinations(range(len(node_ids)), n):
            node_ids_query = [node_ids[i_n] for i_n in i_nodes]
            existence = encoder.get_node_existence_mask(node_ids_query)
            assert np.all(existence == (status_all[:, i_nodes] == Diag.CONFIRMED.value))

            for i_dv in range(status_all.shape[0]):
                existence_i = encoder.get_node_existence_mask(node_ids_query, i_dv=i_dv)
                assert np.all(existence_i == (status_all[i_dv, i_nodes] == Diag.CONFIRMED.value))

    assert_behavior(encoder)



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.matrix import *
from adsg_core.optimization.assign_enc.lazy_encoding import *

__all__ = ['InvalidPatternEncoder', 'PatternEncoderBase']


class InvalidPatternEncoder(RuntimeError):
    pass


class PatternEncoderBase(LazyEncoder):
    """
    Base class for encoders that can only be used for specific connection patterns. The advantage is that encoder is
    very fast, optimal in terms of imputation ratio and distance correlation, and has custom imputation that is better
    and faster than automatic imputation.

    However, it only applies to a limited set of connection settings. When incompatible settings are provided to
    set_settings, an InvalidPatternEncoder exception is raised.
    """
    _allow_random_imputation = False

    def __init__(self, imputer):
        super().__init__(imputer)
        self._settings: Optional[MatrixGenSettings] = None
        self._effective_settings = None
        self._is_transpose = False
        self._transpose_existence_map = {}

    def set_settings(self, settings: MatrixGenSettings):
        # Get effective connection settings and check if all of them are appropriate for this pattern encoder
        try:
            self._is_transpose = False
            self._try_settings(settings)
            self._transpose_existence_map = {}

        except InvalidPatternEncoder:
            # Check if the transpose settings would work (as all patterns are implemented in one way only)
            self._is_transpose = True
            self._try_settings(settings.get_transpose_settings())
            self._transpose_existence_map = {existence: existence.get_transpose() for existence in (
                settings.existence.patterns if settings.existence is not None else [NodeExistence()])}

        # Encode
        super().set_settings(settings)

    def _try_settings(self, settings: MatrixGenSettings):
        self._settings = settings
        self._effective_settings = settings.get_effective_settings()
        if not self._is_compatible_effective(self._effective_settings):
            raise InvalidPatternEncoder(f'Invalid pattern encoder {self!r}')

    def is_compatible(self, settings: MatrixGenSettings) -> bool:
        """Returns whether this encoder is compatible with the provided matrix settings"""
        if self._is_compatible_effective(settings.get_effective_settings()):
            return True
        if self._is_compatible_effective(settings.get_transpose_settings().get_effective_settings()):
            return True
        return False

    def _is_compatible_effective(self, effective_settings) -> bool:
        initialize = True
        for effective_settings, _, _ in effective_settings.values():
            if len(effective_settings.src) == 0 or len(effective_settings.tgt) == 0:
                continue
            if not self._matches_pattern(effective_settings, initialize=initialize):
                return False
            initialize = False
        return True

    def _matches_pattern(self, effective_settings: MatrixGenSettings, initialize: bool) -> bool:
        """Returns whether this encoder can be used for these matrix settings, i.e. whether it matches the pattern"""
        raise NotImplementedError

    def _encode(self, existence: NodeExistence) -> List[DiscreteDV]:
        existence = self._get_existence(existence)
        if existence is None:
            return []

        effective_settings, _, _ = self._effective_settings[existence]
        if len(effective_settings.src) == 0 or len(effective_settings.tgt) == 0:
            return []

        return self._encode_effective(effective_settings, existence)

    def _decode(self, vector: DesignVector, existence: NodeExistence) -> Optional[Tuple[DesignVector, np.ndarray]]:
        existence = self._get_existence(existence)
        if existence is None:
            return vector, self.get_empty_matrix()

        effective_settings, src_map, tgt_map = self._effective_settings[existence]
        if len(effective_settings.src) == 0 or len(effective_settings.tgt) == 0:
            return vector, self.get_empty_matrix()

        vector, matrix = self._decode_effective(vector, effective_settings, existence)
        expanded_matrix = self._settings.expand_effective_matrix(matrix, src_map, tgt_map)

        if self._is_transpose:
            expanded_matrix = expanded_matrix.T

        return vector, expanded_matrix

    def _generate_random_dv_mat(self, n: int, existence: NodeExistence) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        original_existence = existence
        existence = self._get_existence(existence)
        if existence is None:
            return

        effective_settings, src_map, tgt_map = self._effective_settings[existence]
        if len(effective_settings.src) == 0 or len(effective_settings.tgt) == 0:
            _, n_dv, _, _ = self._get_des_vars_n_extra(original_existence)
            return np.zeros((0, n_dv), dtype=int), np.array([self.get_empty_matrix()])[:0, :, :]

        design_vectors, matrices = self._do_generate_random_dv_mat(n, effective_settings, existence)
        expanded_matrices = self._settings.expand_effective_matrix(matrices, src_map, tgt_map)

        if self._is_transpose:
            expanded_matrices = expanded_matrices.transpose((0, 2, 1))
        return design_vectors, expanded_matrices

    def _get_all_design_vectors(self, patterns: List[NodeExistence]) -> Optional[Dict[NodeExistence, np.ndarray]]:
        dv_map = {}
        for existence in patterns:
            original_existence = existence
            existence = self._get_existence(existence)
            if existence is None:
                continue

            design_vars = self._existence_design_vars.get(original_existence, self.design_vars)
            effective_settings, _, _ = self._effective_settings[existence]
            if len(effective_settings.src) == 0 or len(effective_settings.tgt) == 0:
                null_dvs = -np.ones((1, len(design_vars)), dtype=int)
                dv_map[original_existence] = null_dvs
                continue

            design_vectors = self._do_get_all_design_vectors(effective_settings, existence, design_vars)
            if design_vectors.shape[1] != len(design_vars):
                raise RuntimeError(f'Inconsistent nr of design variables: {len(design_vars)}')
            dv_map[original_existence] = design_vectors
        return dv_map

    def _get_existence(self, existence: NodeExistence) -> Optional[NodeExistence]:
        if self._is_transpose:
            if existence not in self._transpose_existence_map:
                return
            existence = self._transpose_existence_map[existence]

        if existence not in self._effective_settings:
            return
        return existence

    def _impute(self, vector, matrix, existence: NodeExistence) -> Tuple[DesignVector, np.ndarray]:
        raise RuntimeError('Pattern encoder should never (automatically) impute!')

    def _encode_effective(self, effective_settings: MatrixGenSettings, existence: NodeExistence) -> List[DiscreteDV]:
        """Encode an effective settings pattern"""
        raise NotImplementedError

    def _decode_effective(self, vector: DesignVector, effective_settings: MatrixGenSettings, existence: NodeExistence) \
            -> Tuple[DesignVector, np.ndarray]:
        """Decode and correct a design vector"""
        raise NotImplementedError

    def _do_generate_random_dv_mat(self, n: int, effective_settings: MatrixGenSettings, existence: NodeExistence) \
            -> Tuple[np.ndarray, np.ndarray]:
        """Generate (n x nx) random vectors and (n x n_src x n_tgt) associated random matrices"""
        raise NotImplementedError

    def _do_get_all_design_vectors(self, effective_settings: MatrixGenSettings, existence: NodeExistence,
                                   design_vars: List[DiscreteDV]) -> np.ndarray:
        """Generate all design vectors (n x nx)"""
        raise NotImplementedError

    def _pattern_name(self) -> str:
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r})'

    def __str__(self):
        return f'{self._pattern_name()} Pattern Encoder'
